@@ -25,7 +25,6 @@ module Client.Commands
   , commandsList
   ) where
 
-import           Client.CApi
 import           Client.Commands.Arguments.Spec
 import           Client.Commands.Arguments.Parser
 import           Client.Commands.Exec
@@ -37,23 +36,25 @@ import           Client.Mask
 import           Client.Message
 import           Client.State
 import           Client.State.Channel
+import           Client.State.DCC
 import qualified Client.State.EditBox as Edit
+import           Client.State.Extensions
 import           Client.State.Focus
 import           Client.State.Network
 import           Client.State.Window
 import           Control.Applicative
+import           Control.Concurrent.Async (cancel)
 import           Control.Exception (displayException, try)
 import           Control.Lens
 import           Control.Monad
 import           Data.Foldable
 import           Data.HashSet (HashSet)
-import           Data.List (nub, (\\))
+import           Data.List (nub, (\\), elem)
 import           Data.List.NonEmpty (NonEmpty((:|)))
 import           Data.List.Split
 import qualified Data.HashMap.Strict as HashMap
 import qualified Data.Map as Map
 import           Data.Maybe (fromMaybe)
-import           Data.Monoid ((<>))
 import           Data.Text (Text)
 import qualified Data.Text as Text
 import           Data.Time
@@ -65,9 +66,9 @@ import           Irc.UserInfo
 import           Irc.Modes
 import           LensUtils
 import           RtsStats (getStats)
+import           System.Directory (doesDirectoryExist)
+import           System.FilePath ((</>))
 import           System.Process
-import           Text.Regex.TDFA
-import           Text.Regex.TDFA.String (compile)
 
 -- | Possible results of running a command
 data CommandResult
@@ -90,7 +91,7 @@ type ChannelCommand a = Identifier {- ^ focused channel -} -> NetworkCommand a
 
 
 -- | Pair of implementations for executing a command and tab completing one.
--- The tab-completion logic is extended with a bool
+-- The tab-completion logic is extended with a Bool
 -- indicating that tab completion should be reversed
 data CommandImpl a
   -- | no requirements
@@ -138,19 +139,30 @@ commandFailureMsg e st =
   return $! CommandFailure $! set clientErrorMsg (Just e) st
 
 -- | Interpret the given chat message or command. Leading @/@ indicates a
--- command. Otherwise if a channel or user query is focused a chat message
--- will be sent.
+-- command. Otherwise if a channel or user query is focused a chat message will
+-- be sent. Leading spaces before the @/@ are ignored when checking for
+-- commands.
 execute ::
-  String {- ^ chat or command -} ->
-  ClientState -> IO CommandResult
+  String           {- ^ chat or command -} ->
+  ClientState      {- ^ client state    -} ->
+  IO CommandResult {- ^ command result  -}
 execute str st =
-  case str of
+  let st' = set clientErrorMsg Nothing st in
+  case dropWhile (' '==) str of
     []          -> commandFailure st
-    '/':command -> executeUserCommand Nothing command st
-    msg         -> executeChat msg st
+    '/':command -> executeUserCommand Nothing command st'
+    _           -> executeChat str st'
 
 -- | Execute command provided by user, resolve aliases if necessary.
-executeUserCommand :: Maybe Text -> String -> ClientState -> IO CommandResult
+--
+-- The last disconnection time is stored in text form and is available
+-- for substitutions in macros. It is only provided when running startup
+-- commands during a reconnect event.
+executeUserCommand ::
+  Maybe Text       {- ^ disconnection time -} ->
+  String           {- ^ command            -} ->
+  ClientState      {- ^ client state       -} ->
+  IO CommandResult {- ^ command result     -}
 executeUserCommand discoTime command st = do
   let key = Text.takeWhile (/=' ') (Text.pack command)
       rest = dropWhile (==' ') (dropWhile (/=' ') command)
@@ -201,14 +213,20 @@ commandExpansion discoTime st v =
 -- | Respond to the TAB key being pressed. This can dispatch to a command
 -- specific completion mode when relevant. Otherwise this will complete
 -- input based on the users of the channel related to the current buffer.
-tabCompletion :: Bool {- ^ reversed -} -> ClientState -> IO CommandResult
+tabCompletion ::
+  Bool             {- ^ reversed       -} ->
+  ClientState      {- ^ client state   -} ->
+  IO CommandResult {- ^ command result -}
 tabCompletion isReversed st =
   case snd $ clientLine st of
     '/':command -> executeCommand (Just isReversed) command st
     _           -> nickTabCompletion isReversed st
 
 -- | Treat the current text input as a chat message and send it.
-executeChat :: String -> ClientState -> IO CommandResult
+executeChat ::
+  String           {- ^ chat message   -} ->
+  ClientState      {- ^ client state   -} ->
+  IO CommandResult {- ^ command result -}
 executeChat msg st =
   case view clientFocus st of
     ChannelFocus network channel
@@ -217,9 +235,7 @@ executeChat msg st =
              let msgTxt = Text.pack $ takeWhile (/='\n') msg
                  tgtTxt = idText channel
 
-             (st1,allow) <- clientPark st $ \ptr ->
-                do let aes = view (clientExtensions . esActive) st
-                   chatExtension ptr network tgtTxt msgTxt aes
+             (st1,allow) <- clientChatExtension network tgtTxt msgTxt st
 
              when allow (sendMsg cs (ircPrivmsg tgtTxt msgTxt))
 
@@ -236,7 +252,11 @@ executeChat msg st =
 -- | Parse and execute the given command. When the first argument is Nothing
 -- the command is executed, otherwise the first argument is the cursor
 -- position for tab-completion
-executeCommand :: Maybe Bool -> String -> ClientState -> IO CommandResult
+executeCommand ::
+  Maybe Bool       {- ^ tab-completion direction -} ->
+  String           {- ^ command                  -} ->
+  ClientState      {- ^ client state             -} ->
+  IO CommandResult {- ^ command result           -}
 
 executeCommand (Just isReversed) _ st
   | Just st' <- commandNameCompletion isReversed st = commandSuccess st'
@@ -312,7 +332,7 @@ commandsList =
 
   , Command
       (pure "reload")
-      (optionalArg (simpleToken "filename"))
+      (optionalArg (simpleToken "[filename]"))
       "Reload the client configuration file.\n\
       \\n\
       \If \^Bfilename\^B is provided it will be used to reload.\n\
@@ -367,7 +387,7 @@ commandsList =
       (remainingArg "arguments")
       "Execute a command synchnonously sending the to a configuration destination.\n\
       \\n\
-      \\^Barguments\^B: [-n network] [-c channel] [-i input] command [command arguments...]\n\
+      \\^Barguments\^B: [-n[network]] [-c[channel]] [-i input] command [arguments...]\n\
       \\n\
       \When \^Binput\^B is specified it is sent to the stdin.\n\
       \\n\
@@ -384,19 +404,27 @@ commandsList =
 
   , Command
       (pure "url")
-      (optionalArg numberArg)
+      optionalNumberArg
       "Open a URL seen in chat.\n\
       \\n\
       \The URL is opened using the executable configured under \^Burl-opener\^B.\n\
       \\n\
-      \When this command is active in the textbox, chat messages are filtered to only show ones with URLs.\n\
+      \When this command is active in the textbox, chat messages are filtered to\
+      \ only show ones with URLs.\n\
       \\n\
-      \When \^Bnumber\^B is omitted it defaults to \^B1\^B. The number selects the URL to open counting back from the most recent.\n"
+      \When \^Bnumber\^B is omitted it defaults to \^B1\^B. The number selects the\
+      \ URL to open counting back from the most recent.\n"
     $ ClientCommand cmdUrl noClientTab
 
+  , Command
+      (pure "cert")
+      (pure ())
+      "Show the TLS certificate for the current connection.\n"
+    $ NetworkCommand cmdCert noNetworkTab
+
   , Command
       (pure "help")
-      (optionalArg (simpleToken "command"))
+      (optionalArg (simpleToken "[command]"))
       "Show command documentation.\n\
       \\n\
       \When \^Bcommand\^B is omitted a list of all commands is displayed.\n\
@@ -479,7 +507,7 @@ commandsList =
 
   [ Command
       (pure "focus")
-      (liftA2 (,) (simpleToken "network") (optionalArg (simpleToken "target")))
+      (liftA2 (,) (simpleToken "network") (optionalArg (simpleToken "[target]")))
       "Change the focused window.\n\
       \\n\
       \When only \^Bnetwork\^B is specified this switches to the network status window.\n\
@@ -549,7 +577,7 @@ commandsList =
 
   , Command
       (pure "clear")
-      (optionalArg (liftA2 (,) (simpleToken "network") (optionalArg (simpleToken "channel"))))
+      (optionalArg (liftA2 (,) (simpleToken "[network]") (optionalArg (simpleToken "[channel]"))))
       "Clear a window.\n\
       \\n\
       \If no arguments are provided the current window is cleared.\n\
@@ -562,7 +590,7 @@ commandsList =
 
   , Command
       (pure "windows")
-      (optionalArg (simpleToken "kind"))
+      (optionalArg (simpleToken "[kind]"))
       "Show a list of all windows with an optional argument to limit the kinds of windows listed.\n\
       \\n\
       \\^Bkind\^O: one of \^Bnetworks\^O, \^Bchannels\^O, \^Busers\^O\n\
@@ -663,27 +691,25 @@ commandsList =
       (remainingArg "regular-expression")
       "Set the persistent regular expression.\n\
       \\n\
-      \Clear the regular expression by calling this without an argument.\n\
-      \\n\
-      \\^B/grep\^O is case-sensitive.\n\
-      \\^B/grepi\^O is case-insensitive.\n"
-    $ ClientCommand (cmdGrep True) simpleClientTab
-
-  , Command
-      (pure "grepi")
-      (remainingArg "regular-expression")
-      "Set the persistent regular expression.\n\
+      \\^BFlags:\^B\n\
+      \    -An  Show n messages after match\n\
+      \    -Bn  Show n messages before match\n\
+      \    -Cn  Show n messages before and after match\n\
+      \    -i   Case insensitive match\n\
+      \    --   Stop processing flags\n\
       \\n\
       \Clear the regular expression by calling this without an argument.\n\
       \\n\
-      \\^B/grep\^O is case-sensitive.\n\
-      \\^B/grepi\^O is case-insensitive.\n"
-    $ ClientCommand (cmdGrep False) simpleClientTab
+      \\^B/grep\^O is case-sensitive.\n"
+    $ ClientCommand cmdGrep simpleClientTab
 
   , Command
       (pure "mentions")
       (pure ())
-      "Show a list of all message that were highlighted as important.\n"
+      "Show a list of all message that were highlighted as important.\n\
+      \\n\
+      \When using \^B/grep\^B the important messages are those matching\n\
+      \the regular expression instead.\n"
     $ ClientCommand cmdMentions noClientTab
 
   ------------------------------------------------------------------------
@@ -692,7 +718,7 @@ commandsList =
 
   [ Command
       ("join" :| ["j"])
-      (liftA2 (,) (simpleToken "channels") (optionalArg (simpleToken "keys")))
+      (liftA2 (,) (simpleToken "channels") (optionalArg (simpleToken "[keys]")))
       "\^BParameters:\^B\n\
       \\n\
       \    channels: Comma-separated list of channels\n\
@@ -905,7 +931,9 @@ commandsList =
       \    Show information about the current channel.\n\
       \    Press ESC to exit the channel info window.\n\
       \\n\
-      \\^BSee also:\^B masks, users\n"
+      \    Information includes topic, creation time, URL, and modes.\n\
+      \\n\
+      \\^BSee also:\^B masks, mode, topic, users\n"
     $ ChannelCommand cmdChannelInfo noChannelTab
 
   , Command
@@ -914,6 +942,17 @@ commandsList =
       "Send a raw IRC command.\n"
     $ NetworkCommand cmdQuote simpleNetworkTab
 
+  , Command
+      (pure "dcc")
+      (liftA2 (,) (optionalArg (simpleToken "[accept|cancel|clear|resume]"))
+                               optionalNumberArg)
+      "Main access to the DCC subsystem with the following subcommands:\n\n\
+       \  /dcc           : Access to a list of pending offer and downloads\n\
+       \  /dcc accept #n : start downloading the #n pending offer\n\
+       \  /dcc resume #n : same as accept but appending to the file on `download-dir`\n\
+       \  /dcc clear  #n : remove the #n offer from the list \n\
+       \  /dcc cancel #n : cancel the download #n \n\n"
+    $ ClientCommand cmdDcc noClientTab
   ------------------------------------------------------------------------
   ] , CommandSection "IRC queries"
   ------------------------------------------------------------------------
@@ -950,7 +989,7 @@ commandsList =
 
   , Command
       (pure "time")
-      (optionalArg (simpleToken "servername"))
+      (optionalArg (simpleToken "[servername]"))
       "Send TIME query to server with given arguments.\n"
     $ NetworkCommand cmdTime simpleNetworkTab
 
@@ -962,22 +1001,22 @@ commandsList =
 
   , Command
       (pure "lusers")
-      (optionalArg (liftA2 (,) (simpleToken "mask") (optionalArg (simpleToken "servername"))))
+      (optionalArg (liftA2 (,) (simpleToken "mask") (optionalArg (simpleToken "[servername]"))))
       "Send LUSERS query to server with given arguments.\n"
     $ NetworkCommand cmdLusers simpleNetworkTab
 
   , Command
-      (pure "motd") (optionalArg (simpleToken "servername"))
+      (pure "motd") (optionalArg (simpleToken "[servername]"))
       "Send MOTD query to server.\n"
     $ NetworkCommand cmdMotd simpleNetworkTab
 
   , Command
-      (pure "admin") (optionalArg (simpleToken "servername"))
+      (pure "admin") (optionalArg (simpleToken "[servername]"))
       "Send ADMIN query to server.\n"
     $ NetworkCommand cmdAdmin simpleNetworkTab
 
   , Command
-      (pure "rules") (optionalArg (simpleToken "servername"))
+      (pure "rules") (optionalArg (simpleToken "[servername]"))
       "Send RULES query to server.\n"
     $ NetworkCommand cmdRules simpleNetworkTab
 
@@ -992,7 +1031,7 @@ commandsList =
     $ NetworkCommand cmdList simpleNetworkTab
 
   , Command
-      (pure "version") (optionalArg (simpleToken "servername"))
+      (pure "version") (optionalArg (simpleToken "[servername]"))
       "Send VERSION query to server.\n"
     $ NetworkCommand cmdVersion simpleNetworkTab
 
@@ -1002,7 +1041,7 @@ commandsList =
 
   [ Command
       (pure "mode")
-      (fromMaybe [] <$> optionalArg (extensionArg "modes" modeParamArgs))
+      (fromMaybe [] <$> optionalArg (extensionArg "[modes]" modeParamArgs))
       "Sets IRC modes.\n\
       \\n\
       \Examples:\n\
@@ -1014,7 +1053,9 @@ commandsList =
       \When executed in a network window, mode changes are applied to your user.\n\
       \When executed in a channel window, mode changes are applied to the channel.\n\
       \\n\
-      \This command has parameter sensitive tab-completion.\n"
+      \This command has parameter sensitive tab-completion.\n\
+      \\n\
+      \See also: /masks /channelinfo\n"
     $ NetworkCommand cmdMode tabMode
 
   , Command
@@ -1028,7 +1069,9 @@ commandsList =
       \\^BI\^B: invite exemptions (op view only)\n\
       \\^Be\^B: ban exemption (op view only)s\n\
       \\n\
-      \To populate the mask lists for the first time use: /mode \^Bmode\^B\n"
+      \To populate the mask lists for the first time use: /mode \^Bmode\^B\n\
+      \\n\
+      \See also: /mode\n"
     $ ChannelCommand cmdMasks noChannelTab
 
   , Command
@@ -1091,7 +1134,7 @@ commandsList =
 
   , Command
       (pure "znc-playback")
-      (optionalArg (liftA2 (,) (simpleToken "time") (optionalArg (simpleToken "date"))))
+      (optionalArg (liftA2 (,) (simpleToken "[time]") (optionalArg (simpleToken "[date]"))))
       "Request playback from the ZNC 'playback' module.\n\
       \\n\
       \\^Btime\^B determines the time to playback since.\n\
@@ -1120,6 +1163,36 @@ commandsList =
       "Kill a client connection to the server.\n"
     $ NetworkCommand cmdKill simpleNetworkTab
 
+  , Command
+      (pure "kline")
+      (liftA3 (,,) (simpleToken "minutes") (simpleToken "user@host") (remainingArg "reason"))
+      "Ban a client from the server.\n"
+    $ NetworkCommand cmdKline simpleNetworkTab
+
+  , Command
+      (pure "unkline")
+      (simpleToken "user@host")
+      "Unban a client from the server.\n"
+    $ NetworkCommand cmdUnkline simpleNetworkTab
+
+  , Command
+      (pure "testline")
+      (simpleToken "[[nick!]user@]host")
+      "Check matching I/K/D lines for a [[nick!]user@]host\n"
+    $ NetworkCommand cmdTestline simpleNetworkTab
+
+  , Command
+      (pure "testmask")
+      (liftA2 (,) (simpleToken "[nick!]user@host") (optionalArg (simpleToken "[gecos]")))
+      "Test how many local and global clients match a mask.\n"
+    $ NetworkCommand cmdTestmask simpleNetworkTab
+
+  , Command
+      (pure "masktrace")
+      (liftA2 (,) (simpleToken "[nick!]user@host") (optionalArg (simpleToken "[gecos]")))
+      "Outputs a list of local users matching the given masks.\n"
+    $ NetworkCommand cmdMasktrace simpleNetworkTab
+
   , Command
       (pure "map")
       (pure ())
@@ -1240,6 +1313,78 @@ cmdQuote cs st rest =
       do sendMsg cs raw
          commandSuccess st
 
+-- | Implementation of @/dcc [(cancel|accept|resume)] [key]
+cmdDcc :: ClientCommand (Maybe String, Maybe Int)
+cmdDcc st (Nothing, Nothing) = commandSuccess (changeSubfocus FocusDCC st)
+cmdDcc st (Just cmd, Just key) = checkAndBranch st cmd key
+cmdDcc st _ = commandFailureMsg "Invalid syntax" st
+
+checkAndBranch :: ClientState -> String -> Int -> IO CommandResult
+checkAndBranch st cmd key
+  | isCancel, NotExist <- curKeyStatus
+      = commandFailureMsg "No such DCC entry" st
+  | isCancel, curKeyStatus == Pending
+      = commandSuccess
+      $ set (clientDCC . dsOffers . ix key . dccStatus) UserKilled st
+  | isCancel, curKeyStatus /= Downloading
+      = commandFailureMsg "Transfer already stopped" st
+  | isCancel = cancel threadId *> commandSuccess st
+
+  | isClear, NotExist <- curKeyStatus
+      = commandFailureMsg "No such DCC entry" st
+  | isClear, curKeyStatus `elem` [Downloading, Pending]
+      = commandFailureMsg "Cancel the download first" st
+  | isClear = commandSuccess
+            $ set (clientDCC . dsOffers    . at key) Nothing
+            $ set (clientDCC . dsTransfers . at key) Nothing st
+
+  | isAcceptOrResume, curKeyStatus `elem` alreadyAcceptedSet
+      = commandFailureMsg "Offer already accepted" st
+  | isAcceptOrResume, NotExist <- curKeyStatus
+      = commandFailureMsg "No such DCC entry" st
+  | isAcceptOrResume
+      = do isDirectory <- doesDirectoryExist downloadPath
+           msize       <- getFileOffset downloadPath
+           case (isDirectory, msize, cmd, mcs) of
+             (True, _, _, _)      -> commandFailureMsg "DCC transfer would overwrite a directory" st
+             (_, Nothing, _, _)   -> acceptOffer -- resume from 0 is accept
+             (_, _, "accept", _)  -> acceptOffer -- overwrite file
+             (_, Just size, "resume", Just cs) -> resumeOffer size cs
+             _ -> commandFailureMsg "Unknown case" st
+
+  | otherwise = commandFailureMsg "Invalid syntax" st
+  where
+    -- General
+    isAcceptOrResume = cmd `elem` ["accept", "resume"]
+    isCancel         = cmd == "cancel"
+    isClear          = cmd == "clear"
+    dccState         = view clientDCC st
+    curKeyStatus     = statusAtKey key dccState
+    alreadyAcceptedSet = [ CorrectlyFinished, UserKilled, LostConnection
+                         , Downloading]
+
+    -- For cancel, other cases handled on the guards
+    threadId = st ^?! clientDCC . dsTransfers . ix key . dtThread . _Just
+
+    -- Common values for resume or accept
+    Just offer   = view (clientDCC . dsOffers . at key) st -- guarded exist
+    updChan      = view clientDCCUpdates st
+    downloadDir  = view (clientConfig . configDownloadDir) st
+    downloadPath = downloadDir </> _dccFileName offer
+    mcs          = preview (clientConnection (_dccNetwork offer)) st
+
+    -- Actual workhorses for the commands
+    acceptOffer =
+        do newDCCState <- supervisedDownload downloadDir key updChan dccState
+           commandSuccess (set clientDCC newDCCState st)
+
+    resumeOffer size cs =
+        let newOffer = offer { _dccOffset = size }
+            (target, txt) = resumeMsg size newOffer
+            st' = set (clientDCC . dsOffers . at key) (Just newOffer) st
+        in cmdCtcp cs st' (target, "DCC", txt)
+
+
 -- | Implementation of @/me@
 cmdMe :: ChannelCommand String
 cmdMe channelId cs st rest =
@@ -1335,7 +1480,7 @@ cmdConnect :: ClientCommand String
 cmdConnect st networkStr =
   do -- abort any existing connection before connecting
      let network = Text.pack networkStr
-     st' <- addConnection 0 Nothing network =<< abortNetwork network st
+     st' <- addConnection 0 Nothing Nothing network =<< abortNetwork network st
      commandSuccess
        $ changeFocus (NetworkFocus network) st'
 
@@ -1544,7 +1689,7 @@ tabConnect :: Bool -> ClientCommand String
 tabConnect isReversed st _ =
   simpleTabCompletion plainWordCompleteMode [] networks isReversed st
   where
-    networks = views clientNetworkMap               HashMap.keys st
+    networks = views clientConnections              HashMap.keys st
             ++ views (clientConfig . configServers) HashMap.keys st
 
 
@@ -1554,7 +1699,7 @@ tabFocus :: Bool -> ClientCommand String
 tabFocus isReversed st _ =
   simpleTabCompletion plainWordCompleteMode [] completions isReversed st
   where
-    networks   = map mkId $ HashMap.keys $ view clientNetworkMap st
+    networks   = map mkId $ HashMap.keys $ view clientConnections st
     params     = words $ uncurry take $ clientLine st
 
     completions =
@@ -1651,6 +1796,31 @@ cmdKill cs st (client,rest) =
   do sendMsg cs (ircKill (Text.pack client) (Text.pack rest))
      commandSuccess st
 
+cmdKline :: NetworkCommand (String, String, String)
+cmdKline cs st (minutes, mask, reason) =
+  do sendMsg cs (ircKline (Text.pack minutes) (Text.pack mask) (Text.pack reason))
+     commandSuccess st
+
+cmdUnkline :: NetworkCommand String
+cmdUnkline cs st mask =
+  do sendMsg cs (ircUnkline (Text.pack mask))
+     commandSuccess st
+
+cmdTestline :: NetworkCommand String
+cmdTestline cs st mask =
+  do sendMsg cs (ircTestline (Text.pack mask))
+     commandSuccess st
+
+cmdTestmask :: NetworkCommand (String, Maybe String)
+cmdTestmask cs st (mask, gecos) =
+  do sendMsg cs (ircTestmask (Text.pack mask) (maybe "" Text.pack gecos))
+     commandSuccess st
+
+cmdMasktrace :: NetworkCommand (String, Maybe String)
+cmdMasktrace cs st (mask, gecos) =
+  do sendMsg cs (ircMasktrace (Text.pack mask) (maybe "*" Text.pack gecos))
+     commandSuccess st
+
 cmdAway :: NetworkCommand String
 cmdAway cs st rest =
   do sendMsg cs (ircAway (Text.pack rest))
@@ -1748,9 +1918,9 @@ cmdInvite channelId cs st nick =
 
 commandSuccessUpdateCS :: NetworkState -> ClientState -> IO CommandResult
 commandSuccessUpdateCS cs st =
-  do let networkId = view csNetworkId cs
+  do let network = view csNetwork cs
      commandSuccess
-       $ setStrict (clientConnections . ix networkId) cs st
+       $ setStrict (clientConnection network) cs st
 
 cmdTopic :: ChannelCommand String
 cmdTopic channelId cs st rest =
@@ -1814,8 +1984,8 @@ cmdKickBan channelId cs st (who,reason) =
 computeBanUserInfo :: Identifier -> NetworkState    -> UserInfo
 computeBanUserInfo who cs =
   case view (csUser who) cs of
-    Nothing                   -> UserInfo who "*" "*"
-    Just (UserAndHost _ host) -> UserInfo "*" "*" host
+    Nothing                     -> UserInfo who "*" "*"
+    Just (UserAndHost _ host _) -> UserInfo "*" "*" host
 
 cmdRemove :: ChannelCommand (String, String)
 cmdRemove channelId cs st (who,reason) =
@@ -1912,7 +2082,7 @@ cmdReconnect st _
   | Just network <- views clientFocus focusNetwork st =
 
       do tm <- getCurrentTime
-         st' <- addConnection 0 (Just tm) network =<< abortNetwork network st
+         st' <- addConnection 0 (Just tm) Nothing network =<< abortNetwork network st
          commandSuccess
            $ changeFocus (NetworkFocus network) st'
 
@@ -1947,14 +2117,14 @@ tabIgnore isReversed st _ =
 -- Attempt to reload the configuration file
 cmdReload :: ClientCommand (Maybe String)
 cmdReload st mbPath =
-  do let path   = mbPath <|> view clientConfigPath st
+  do let path = mbPath <|> Just (view clientConfigPath st)
      res <- loadConfiguration path
      case res of
        Left e -> commandFailureMsg (describeProblem e) st
-       Right cfg ->
+       Right (path',cfg) ->
          do st1 <- clientStartExtensions
                  $ set clientConfig cfg
-                 $ set clientConfigPath mbPath st
+                 $ set clientConfigPath path' st
             commandSuccess st1
 
   where
@@ -2125,7 +2295,7 @@ computeModeCompletion pol mode channel cs st
     usermasks =
       [ mkId ("*!*@" <> host)
         | nick <- HashMap.keys (view (csChannels . ix channel . chanUsers) cs)
-        , UserAndHost _ host <- toListOf (csUsers . ix nick) cs
+        , UserAndHost _ host _ <- toListOf (csUsers . ix nick) cs
         ]
 
 -- | Predicate for mode commands that can be performed without ops
@@ -2158,13 +2328,10 @@ nickTabCompletion isReversed st =
 
 cmdExtension :: ClientCommand (String, String)
 cmdExtension st (name,command) =
-  case find (\ae -> aeName ae == Text.pack name)
-            (view (clientExtensions . esActive) st) of
-        Nothing -> commandFailureMsg "unknown extension" st
-        Just ae ->
-          do (st',_) <- clientPark st $ \ptr ->
-                          commandExtension ptr (Text.pack command) ae
-             commandSuccess st'
+  do res <- clientCommandExtension (Text.pack name) (Text.pack command) st
+     case res of
+       Nothing  -> commandFailureMsg "unknown extension" st
+       Just st' -> commandSuccess st'
 
 -- | Implementation of @/exec@ command.
 cmdExec :: ClientCommand String
@@ -2185,19 +2352,36 @@ cmdExec st rest =
     buildTransmitter now ec =
       case (Text.pack <$> view execOutputNetwork ec,
             Text.pack <$> view execOutputChannel ec) of
-        (Nothing, Nothing) -> Right (sendToClient now)
-        (Just network, Nothing) ->
+
+        (Unspecified, Unspecified) -> Right (sendToClient now)
+
+        (Specified network, Specified channel) ->
           case preview (clientConnection network) st of
             Nothing -> Left ["Unknown network"]
-            Just cs -> Right (sendToNetwork now cs)
-        (Nothing , Just channel) ->
+            Just cs -> Right (sendToChannel cs channel)
+
+        (_ , Specified channel) ->
           case currentNetworkState of
             Nothing -> Left ["No current network"]
             Just cs -> Right (sendToChannel cs channel)
-        (Just network, Just channel) ->
+
+        (Specified network, _) ->
           case preview (clientConnection network) st of
             Nothing -> Left ["Unknown network"]
-            Just cs -> Right (sendToChannel cs channel)
+            Just cs -> Right (sendToNetwork now cs)
+
+        (_, Current) ->
+          case currentNetworkState of
+            Nothing -> Left ["No current network"]
+            Just cs ->
+              case view clientFocus st of
+                ChannelFocus _ channel -> Right (sendToChannel cs (idText channel))
+                _                      -> Left ["No current channel"]
+
+        (Current, _) ->
+          case currentNetworkState of
+            Nothing -> Left ["No current network"]
+            Just cs -> Right (sendToNetwork now cs)
 
     sendToClient now msgs = commandSuccess $! foldl' (recordSuccess now) st msgs
 
@@ -2206,7 +2390,7 @@ cmdExec st rest =
       foldM (\st1 msg ->
            case parseRawIrcMsg msg of
              Nothing ->
-               return $! recordError now st1 ("Bad raw message: " <> msg)
+               return $! recordError now "" ("Bad raw message: " <> msg) st1
              Just raw ->
                do sendMsg cs raw
                   return st1) st msgs
@@ -2225,15 +2409,7 @@ cmdExec st rest =
          preview (clientConnection network) st
 
     failure now es =
-      commandFailure $! foldl' (recordError now) st (map Text.pack es)
-
-recordError :: ZonedTime -> ClientState -> Text -> ClientState
-recordError now ste e =
-  recordNetworkMessage ClientMessage
-    { _msgTime    = now
-    , _msgBody    = ErrorBody e
-    , _msgNetwork = ""
-    } ste
+      commandFailure $! foldl' (flip (recordError now "")) st (map Text.pack es)
 
 recordSuccess :: ZonedTime -> ClientState -> Text -> ClientState
 recordSuccess now ste m =
@@ -2267,18 +2443,19 @@ openUrl opener url st =
        Left e  -> commandFailureMsg (Text.pack (displayException (e :: IOError))) st
        Right{} -> commandSuccess st
 
--- | Implementation of @/grep@ and @/grepi@
-cmdGrep ::
-  Bool {- ^ case sensitive -} ->
-  ClientCommand String
-cmdGrep sensitive st str
+-- | Implementation of @/grep@
+cmdGrep :: ClientCommand String
+cmdGrep st str
   | null str  = commandSuccess (set clientRegex Nothing st)
   | otherwise =
-      case compile defaultCompOpt{caseSensitive=sensitive} defaultExecOpt str of
-        Left e -> commandFailureMsg (Text.pack e) st
-        Right r -> commandSuccess (set clientRegex (Just r) st)
+      case buildMatcher str of
+        Nothing -> commandFailureMsg "bad grep" st
+        Just  r -> commandSuccess (set clientRegex (Just r) st)
 
 cmdOper :: NetworkCommand (String, String)
 cmdOper cs st (user, pass) =
   do sendMsg cs (ircOper (Text.pack user) (Text.pack pass))
      commandSuccess st
+
+cmdCert :: NetworkCommand ()
+cmdCert _ st _ = commandSuccess (changeSubfocus FocusCert st)
