@@ -3,6 +3,7 @@
 {-# LANGUAGE RecordWildCards   #-}
 {-# LANGUAGE TemplateHaskell   #-}
 {-# LANGUAGE ApplicativeDo     #-}
+{-# LANGUAGE RankNTypes        #-}
 
 {-|
 Module      : Client.Configuration
@@ -21,6 +22,7 @@ module Client.Configuration
   , ConfigurationFailure(..)
   , LayoutMode(..)
   , PaddingMode(..)
+  , ExtensionConfiguration(..)
 
   -- * Lenses
   , configDefaults
@@ -28,6 +30,7 @@ module Client.Configuration
   , configPalette
   , configWindowNames
   , configNickPadding
+  , configDownloadDir
   , configMacros
   , configExtensions
   , configExtraHighlights
@@ -41,6 +44,10 @@ module Client.Configuration
   , configShowPing
   , configJumpModifier
 
+  , extensionPath
+  , extensionRtldFlags
+  , extensionArgs
+
   -- * Loading configuration
   , loadConfiguration
 
@@ -49,6 +56,11 @@ module Client.Configuration
 
   -- * Specification
   , configurationSpec
+
+  -- * FilePath resolution
+  , FilePathContext
+  , newFilePathContext
+  , resolveFilePath
   ) where
 
 import           Client.Commands.Interpolation
@@ -61,8 +73,9 @@ import           Client.Image.Palette
 import           Config
 import           Config.Schema
 import           Control.Exception
+import           Control.Monad                       (unless)
 import           Control.Lens                        hiding (List)
-import           Data.Foldable                       (toList, find)
+import           Data.Foldable                       (toList)
 import           Data.Functor.Alt                    ((<!>))
 import           Data.HashMap.Strict                 (HashMap)
 import qualified Data.HashMap.Strict                 as HashMap
@@ -70,16 +83,18 @@ import           Data.HashSet                        (HashSet)
 import qualified Data.HashSet                        as HashSet
 import qualified Data.List.NonEmpty                  as NonEmpty
 import           Data.Maybe
-import           Data.Monoid                         (Endo(..), (<>))
+import           Data.Monoid                         (Endo(..))
 import           Data.Text                           (Text)
 import qualified Data.Text                           as Text
 import qualified Data.Text.IO                        as Text
 import qualified Data.Vector                         as Vector
 import           Graphics.Vty.Input.Events (Modifier(..), Key(..))
+import           Graphics.Vty.Attributes             (Attr)
 import           Irc.Identifier                      (Identifier)
 import           System.Directory
 import           System.FilePath
 import           System.IO.Error
+import           System.Posix.DynamicLinker          (RTLDFlags(..))
 
 -- | Top-level client configuration information. When connecting to a
 -- server configuration from '_configServers' is used where possible,
@@ -87,12 +102,13 @@ import           System.IO.Error
 data Configuration = Configuration
   { _configDefaults        :: ServerSettings -- ^ Default connection settings
   , _configServers         :: (HashMap Text ServerSettings) -- ^ Host-specific settings
-  , _configPalette         :: Palette
+  , _configPalette         :: Palette -- ^ User-customized color palette
   , _configWindowNames     :: Text -- ^ Names of windows, used when alt-jumping)
   , _configExtraHighlights :: HashSet Identifier -- ^ Extra highlight nicks/terms
   , _configNickPadding     :: PaddingMode -- ^ Padding of nicks in messages
+  , _configDownloadDir     :: FilePath -- ^ Directory for downloads, default to HOME
   , _configMacros          :: Recognizer Macro -- ^ command macros
-  , _configExtensions      :: [FilePath] -- ^ paths to shared library
+  , _configExtensions      :: [ExtensionConfiguration] -- ^ extensions to load
   , _configUrlOpener       :: Maybe FilePath -- ^ paths to url opening executable
   , _configIgnores         :: [Text] -- ^ initial ignore mask list
   , _configActivityBar     :: Bool -- ^ initially visibility of the activity bar
@@ -119,7 +135,6 @@ data LayoutMode
   | TwoColumn
   deriving Show
 
-makeLenses ''Configuration
 
 -- | Failure cases when loading a configuration file.
 data ConfigurationFailure
@@ -137,16 +152,31 @@ data ConfigurationFailure
 -- | default instance
 instance Exception ConfigurationFailure
 
+-- | Configuration information for run-time loaded dynamic
+-- library extensions.
+data ExtensionConfiguration = ExtensionConfiguration
+  { _extensionPath      :: FilePath -- ^ path to shared object
+  , _extensionRtldFlags :: [RTLDFlags] -- ^ dynamic linker flags
+  , _extensionArgs      :: [Text] -- ^ arguments to the extension on startup
+  }
+  deriving Show
+
+makeLenses ''Configuration
+makeLenses ''ExtensionConfiguration
+
+-- | The default client behavior for naming windows is to use the first two
+-- rows of a QWERTY keyboard followed by the first two rows combined with
+-- SHIFT.
 defaultWindowNames :: Text
 defaultWindowNames = "1234567890qwertyuiop!@#$%^&*()QWERTYUIOP"
 
--- | Uses 'getAppUserDataDirectory' to find @.glirc/config@
+-- | Uses 'getAppUserDataDirectory' to find @~/.glirc/config@
 getOldConfigPath :: IO FilePath
 getOldConfigPath =
   do dir <- getAppUserDataDirectory "glirc"
      return (dir </> "config")
 
--- | Uses 'getXdgDirectory' 'XdgConfig' to find @.config/glirc/config@
+-- | Uses 'getXdgDirectory' 'XdgConfig' to find @~/.config/glirc/config@
 getNewConfigPath :: IO FilePath
 getNewConfigPath =
   do dir <- getXdgDirectory XdgConfig "glirc"
@@ -194,12 +224,16 @@ readConfigurationFile mbPath =
 
 
 -- | Load the configuration file defaulting to @~/.glirc/config@.
+--
+-- Given configuration path is optional and actual path used will
+-- be returns on success
 loadConfiguration ::
   Maybe FilePath {- ^ path to configuration file -} ->
-  IO (Either ConfigurationFailure Configuration)
+  IO (Either ConfigurationFailure (FilePath, Configuration))
 loadConfiguration mbPath = try $
   do (path,txt) <- readConfigurationFile mbPath
      def  <- loadDefaultServerSettings
+     home <- getHomeDirectory
 
      rawcfg <-
        case parse txt of
@@ -207,32 +241,13 @@ loadConfiguration mbPath = try $
          Right rawcfg -> return rawcfg
 
      case loadValue configurationSpec rawcfg of
-       Left es -> throwIO
-                $ ConfigurationMalformed path
-                $ Text.unpack
-                $ Text.unlines
-                $ map explainLoadError (toList es)
-       Right cfg -> resolvePaths path (cfg def)
-
-
-explainLoadError :: LoadError Position -> Text
-explainLoadError (LoadError pos path problem) =
-  Text.concat [ positionText, " at ", pathText, ": ", problemText]
-
-  where
-    positionText =
-     Text.unwords ["line"  , Text.pack (show (posLine   pos)),
-                   "column", Text.pack (show (posColumn pos))]
-
-    pathText
-      | null path = "top-level"
-      | otherwise = Text.intercalate ":" path
-
-    problemText =
-      case problem of
-        UnusedSection  s -> "unknown section `"          <> s <> "`"
-        MissingSection s -> "missing required section `" <> s <> "`"
-        SpecMismatch   t -> "expected "                  <> t
+       Left e -> throwIO
+               $ ConfigurationMalformed path
+               $ displayException e
+       Right cfg ->
+         do cfg' <- resolvePaths path (cfg def home)
+                    >>= validateDirectories path
+            return (path, cfg')
 
 
 -- | Resolve all the potentially relative file paths in the configuration file
@@ -244,13 +259,29 @@ resolvePaths file cfg =
                                 . over (ssTlsServerCert . mapped) res
                                 . over (ssSaslEcdsaFile . mapped) res
                                 . over (ssLogDir        . mapped) res
-     return $! over (configExtensions . mapped) res
+     return $! over (configExtensions . mapped . extensionPath) res
              . over (configServers    . mapped) resolveServerFilePaths
+             . over configDownloadDir res
              $ cfg
 
+-- | Check if the `download-dir` is actually a directory and writeable,
+--   throw a ConfigurationMalformed exception if it isn't.
+validateDirectories :: FilePath -> Configuration -> IO Configuration
+validateDirectories cfgPath cfg =
+  do isDir       <- doesDirectoryExist downloadPath
+     unless isDir $ throwIO (ConfigurationMalformed cfgPath noDirMsg)
+     isWriteable <- writable <$> getPermissions downloadPath
+     unless isWriteable
+       $ throwIO (ConfigurationMalformed cfgPath noWriteableMsg)
+     return cfg
+  where
+    downloadPath = view configDownloadDir cfg
+    noDirMsg = "The download-dir section doesn't point to a directory."
+    noWriteableMsg = "The download-dir doesn't point to a writeable directory."
+
 configurationSpec ::
-  ValueSpecs (ServerSettings -> Configuration)
-configurationSpec = sectionsSpec "" $
+  ValueSpec (ServerSettings -> FilePath -> Configuration)
+configurationSpec = sectionsSpec "config-file" $
 
   do let sec' def name spec info = fromMaybe def <$> optSection' name spec info
          identifierSetSpec       = HashSet.fromList <$> listSpec identifierSpec
@@ -261,21 +292,21 @@ configurationSpec = sectionsSpec "" $
                                "Configuration parameters for IRC servers"
      _configPalette         <- sec' defaultPalette "palette" paletteSpec
                                "Customize the client color choices"
-     _configWindowNames     <- sec' defaultWindowNames "window-names" valuesSpec
+     _configWindowNames     <- sec' defaultWindowNames "window-names" anySpec
                                "Window names to use for quick jumping with jump-modifier key"
      _configJumpModifier    <- sec' [MMeta] "jump-modifier" modifierSpec
                                "Modifier used to jump to a window by name. Defaults to `meta`."
      _configMacros          <- sec' mempty "macros" macroMapSpec
                                "Programmable macro commands"
-     _configExtensions      <- sec' [] "extensions" (listSpec stringSpec)
-                               "Filenames of extension libraries to load at startup"
+     _configExtensions      <- sec' [] "extensions" (listSpec extensionSpec)
+                               "extension libraries to load at startup"
      _configUrlOpener       <- optSection' "url-opener" stringSpec
                                "External command used by /url command"
      _configExtraHighlights <- sec' mempty "extra-highlights" identifierSetSpec
                                "Extra words to highlight in chat messages"
      _configNickPadding     <- sec' NoPadding "nick-padding" nickPaddingSpec
                                "Amount of space to reserve for nicknames in chat messages"
-     _configIgnores         <- sec' [] "ignores" valuesSpec
+     _configIgnores         <- sec' [] "ignores" anySpec
                                "Set of nicknames to ignore on startup"
      _configActivityBar     <- sec' False  "activity-bar" yesOrNoSpec
                                "Show channel names and message counts for activity on\
@@ -290,10 +321,13 @@ configurationSpec = sectionsSpec "" $
                                "Initial setting for window layout"
      _configShowPing        <- sec' True "show-ping" yesOrNoSpec
                                "Initial setting for visibility of ping times"
-     return (\def ->
+     maybeDownloadDir       <- optSection' "download-dir" stringSpec
+                               "Path to DCC download directoy. Defaults to home directory."
+     return (\def home ->
              let _configDefaults = ssDefUpdate def
                  _configServers  = buildServerMap _configDefaults ssUpdates
                  _configKeyMap   = foldl (\acc f -> f acc) initialKeyMap bindings
+                 _configDownloadDir = fromMaybe home maybeDownloadDir
              in Configuration{..})
 
 -- | The default nick padding side if padding is going to be used
@@ -307,7 +341,7 @@ defaultPaddingSide = RightPadding
 -- > nick-padding:
 -- >   side: right
 -- >   width: 16
-nickPaddingSpec :: ValueSpecs PaddingMode
+nickPaddingSpec :: ValueSpec PaddingMode
 nickPaddingSpec = defaultPaddingSide <$> nonnegativeSpec <!> fullNickPaddingSpec
 
 -- | Full nick padding specification:
@@ -315,7 +349,7 @@ nickPaddingSpec = defaultPaddingSide <$> nonnegativeSpec <!> fullNickPaddingSpec
 -- > nick-padding:
 -- >   side: left
 -- >   width: 15
-fullNickPaddingSpec :: ValueSpecs PaddingMode
+fullNickPaddingSpec :: ValueSpec PaddingMode
 fullNickPaddingSpec = sectionsSpec "nick-padding" (sideSec <*> amtSec)
   where
     sideSpec = LeftPadding  <$ atomSpec "left" <!>
@@ -327,7 +361,9 @@ fullNickPaddingSpec = sectionsSpec "nick-padding" (sideSec <*> amtSec)
     amtSec  = reqSection' "width" nonnegativeSpec "Field width"
 
 
-modifierSpec :: ValueSpecs [Modifier]
+-- | Parse either a single modifier key or a list of modifier keys:
+-- @meta@, @alt@, @ctrl@
+modifierSpec :: ValueSpec [Modifier]
 modifierSpec = toList <$> oneOrNonemptySpec modifier1Spec
   where
     modifier1Spec = namedSpec "modifier"
@@ -335,14 +371,23 @@ modifierSpec = toList <$> oneOrNonemptySpec modifier1Spec
                 <!> MAlt  <$ atomSpec "alt"
                 <!> MCtrl <$ atomSpec "ctrl"
 
-layoutSpec :: ValueSpecs LayoutMode
+-- | Parse either @one-column@ or @two-column@ and return the corresponding
+-- 'LayoutMode' value.
+layoutSpec :: ValueSpec LayoutMode
 layoutSpec = OneColumn <$ atomSpec "one-column"
          <!> TwoColumn <$ atomSpec "two-column"
 
-keyBindingSpec :: ValueSpecs (KeyMap -> KeyMap)
+-- | Parse a single key binding. This can be an action binding, command
+-- binding, or an unbinding specification.
+keyBindingSpec :: ValueSpec (KeyMap -> KeyMap)
 keyBindingSpec = actBindingSpec <!> cmdBindingSpec <!> unbindingSpec
 
-actBindingSpec :: ValueSpecs (KeyMap -> KeyMap)
+-- | Parse a single action key binding. Action bindings are a map specifying
+-- a binding using 'keySpec' and an action:
+--
+-- > bind: "M-a"
+-- > action: jump-to-activity
+actBindingSpec :: ValueSpec (KeyMap -> KeyMap)
 actBindingSpec = sectionsSpec "action-binding" $
   do ~(m,k) <- reqSection' "bind" keySpec
                "Key to be bound (e.g. a, C-b, M-c C-M-d)"
@@ -350,7 +395,7 @@ actBindingSpec = sectionsSpec "action-binding" $
                "Action name (see `/keymap`)"
      return (addKeyBinding m k a)
 
-cmdBindingSpec :: ValueSpecs (KeyMap -> KeyMap)
+cmdBindingSpec :: ValueSpec (KeyMap -> KeyMap)
 cmdBindingSpec = sectionsSpec "command-binding" $
   do ~(m,k) <- reqSection' "bind" keySpec
                "Key to be bound (e.g. a, C-b, M-c C-M-d)"
@@ -358,7 +403,7 @@ cmdBindingSpec = sectionsSpec "command-binding" $
                "Client command to execute (exclude leading `/`)"
      return (addKeyBinding m k (ActCommand cmd))
 
-unbindingSpec :: ValueSpecs (KeyMap -> KeyMap)
+unbindingSpec :: ValueSpec (KeyMap -> KeyMap)
 unbindingSpec = sectionsSpec "remove-binding" $
   do ~(m,k) <- reqSection' "unbind" keySpec
                "Key to be unbound (e.g. a, C-b, M-c C-M-d)"
@@ -366,26 +411,91 @@ unbindingSpec = sectionsSpec "remove-binding" $
 
 
 -- | Custom configuration specification for emacs-style key descriptions
-keySpec :: ValueSpecs ([Modifier], Key)
-keySpec = customSpec "emacs-key" stringSpec parseKey
+keySpec :: ValueSpec ([Modifier], Key)
+keySpec = customSpec "emacs-key" stringSpec
+        $ \key -> case parseKey key of
+                    Nothing -> Left "unknown key"
+                    Just x  -> Right x
 
 
-nonnegativeSpec :: (Ord a, Num a) => ValueSpecs a
-nonnegativeSpec = customSpec "non-negative" numSpec $ \x -> find (0 <=) [x]
+nonnegativeSpec :: (Ord a, Num a) => ValueSpec a
+nonnegativeSpec = customSpec "non-negative" numSpec
+                $ \x -> if x < 0 then Left "negative number"
+                                 else Right x
 
 
-paletteSpec :: ValueSpecs Palette
+paletteSpec :: ValueSpec Palette
 paletteSpec = sectionsSpec "palette" $
   (ala Endo (foldMap . foldMap) ?? defaultPalette) <$> sequenceA fields
 
   where
-    nickColorsSpec = set palNicks . Vector.fromList . NonEmpty.toList <$> nonemptySpec attrSpec
+    nickColorsSpec :: ValueSpec (Palette -> Palette)
+    nickColorsSpec = set palNicks . Vector.fromList . NonEmpty.toList
+                 <$> nonemptySpec attrSpec
 
-    fields :: [SectionSpecs (Maybe (Palette -> Palette))]
+    modeColorsSpec :: Lens' Palette (HashMap Char Attr) -> ValueSpec (Palette -> Palette)
+    modeColorsSpec l
+      = fmap (set l)
+      $ customSpec "modes" (assocSpec attrSpec)
+      $ fmap HashMap.fromList
+      . traverse (\(mode, attr) ->
+          case Text.unpack mode of
+            [m] -> Right (m, attr)
+            _   -> Left "expected single letter")
+
+    fields :: [SectionsSpec (Maybe (Palette -> Palette))]
     fields = optSection' "nick-colors" nickColorsSpec
              "Colors used to highlight nicknames"
+
+           : optSection' "cmodes" (modeColorsSpec palCModes)
+             "Colors used to highlight channel modes"
+
+           : optSection' "umodes" (modeColorsSpec palUModes)
+             "Colors used to highlight user modes"
+
+           : optSection' "snomask" (modeColorsSpec palSnomask)
+             "Colors used to highlight server notice mask"
+
            : [ optSection' lbl (set l <$> attrSpec) "" | (lbl, Lens l) <- paletteMap ]
 
+extensionSpec :: ValueSpec ExtensionConfiguration
+extensionSpec = simpleExtensionSpec <!> fullExtensionSpec
+
+-- | Default dynamic linker flags: @RTLD_LOCAL@ and @RTLD_NOW@
+defaultRtldFlags :: [RTLDFlags]
+defaultRtldFlags = [RTLD_LOCAL, RTLD_NOW]
+
+-- | Given only a filepath build an extension configuration that
+-- loads the extension using the 'defaultRtldFlags' and no arguments.
+simpleExtensionSpec :: ValueSpec ExtensionConfiguration
+simpleExtensionSpec =
+  do _extensionPath <- stringSpec
+     pure ExtensionConfiguration
+       { _extensionRtldFlags = defaultRtldFlags
+       , _extensionArgs      = []
+       , .. }
+
+-- | Full extension configuration allows the RTLD flags to be manually
+-- specified. This can be useful if the extension defines symbols that
+-- need to be visible to libraries that the extension is linked against.
+fullExtensionSpec :: ValueSpec ExtensionConfiguration
+fullExtensionSpec =
+  sectionsSpec "extension" $
+  do _extensionPath      <- reqSection' "path"       stringSpec
+                            "Path to shared object"
+     _extensionRtldFlags <- fromMaybe defaultRtldFlags <$>
+                            optSection' "rtld-flags" (listSpec rtldFlagSpec)
+                            "Runtime dynamic linker flags"
+     _extensionArgs      <- fromMaybe [] <$> optSection "args"
+                            "Extension-specific configuration arguments"
+     pure ExtensionConfiguration {..}
+
+rtldFlagSpec :: ValueSpec RTLDFlags
+rtldFlagSpec = namedSpec "rtld-flag"
+             $ RTLD_LOCAL  <$ atomSpec "local"
+           <!> RTLD_GLOBAL <$ atomSpec "global"
+           <!> RTLD_NOW    <$ atomSpec "now"
+           <!> RTLD_LAZY   <$ atomSpec "lazy"
 
 buildServerMap ::
   ServerSettings {- ^ defaults -} ->
