@@ -48,12 +48,16 @@ module Client.Configuration.ServerSettings
   , ssNickCompletion
   , ssLogDir
   , ssProtocolFamily
+  , ssSts
+  , ssTlsPubkeyFingerprint
+  , ssTlsCertFingerprint
 
   -- * Load function
   , loadDefaultServerSettings
 
   -- * TLS settings
   , UseTls(..)
+  , Fingerprint(..)
 
   ) where
 
@@ -62,14 +66,18 @@ import           Client.Commands.WordCompletion
 import           Client.Configuration.Macros (macroCommandSpec)
 import           Config.Schema.Spec
 import           Control.Lens
+import qualified Data.ByteString as B
 import           Data.Functor.Alt                    ((<!>))
 import           Data.List.NonEmpty (NonEmpty)
+import           Data.ByteString (ByteString)
 import           Data.Maybe (fromMaybe)
 import           Data.Monoid
 import           Data.Text (Text)
+import           Data.List.Split (chunksOf, splitOn)
 import qualified Data.Text as Text
 import           Irc.Identifier (Identifier, mkId)
 import           Network.Socket (HostName, PortNumber, Family(..))
+import           Numeric (readHex)
 import           System.Environment
 
 -- | Static server-level settings
@@ -98,10 +106,13 @@ data ServerSettings = ServerSettings
   , _ssMessageHooks     :: ![Text] -- ^ Initial message hooks
   , _ssName             :: !(Maybe Text) -- ^ The name referencing the server in commands
   , _ssReconnectAttempts:: !Int -- ^ The number of reconnect attempts to make on error
-  , _ssAutoconnect      :: Bool -- ^ Connect to this network on server startup
+  , _ssAutoconnect      :: !Bool -- ^ Connect to this network on server startup
   , _ssNickCompletion   :: WordCompletionMode -- ^ Nick completion mode for this server
   , _ssLogDir           :: Maybe FilePath -- ^ Directory to save logs of chat
   , _ssProtocolFamily   :: Maybe Family -- ^ Protocol family to connect with
+  , _ssSts              :: !Bool -- ^ Honor STS policies when true
+  , _ssTlsPubkeyFingerprint :: !(Maybe Fingerprint) -- ^ optional acceptable public key fingerprint
+  , _ssTlsCertFingerprint   :: !(Maybe Fingerprint) -- ^ optional acceptable certificate fingerprint
   }
   deriving Show
 
@@ -112,6 +123,13 @@ data UseTls
   | UseInsecure    -- ^ Plain connection
   deriving Show
 
+-- | Fingerprint used to validate server certificates.
+data Fingerprint
+  = FingerprintSha1   ByteString -- ^ SHA-1 fingerprint
+  | FingerprintSha256 ByteString -- ^ SHA-2 256-bit fingerprint
+  | FingerprintSha512 ByteString -- ^ SHA-2 512-bit fingerprint
+  deriving Show
+
 makeLenses ''ServerSettings
 
 -- | Load the defaults for server settings based on the environment
@@ -151,9 +169,12 @@ loadDefaultServerSettings =
        , _ssNickCompletion   = defaultNickWordCompleteMode
        , _ssLogDir           = Nothing
        , _ssProtocolFamily   = Nothing
+       , _ssSts              = True
+       , _ssTlsPubkeyFingerprint = Nothing
+       , _ssTlsCertFingerprint   = Nothing
        }
 
-serverSpec :: ValueSpecs (ServerSettings -> ServerSettings)
+serverSpec :: ValueSpec (ServerSettings -> ServerSettings)
 serverSpec = sectionsSpec "server-settings" $
   composeMaybe <$> sequenceA settings
   where
@@ -170,9 +191,9 @@ serverSpec = sectionsSpec "server-settings" $
       $ set l . Just <$> s <!>
         set l Nothing <$ atomSpec "clear"
 
-    settings :: [SectionSpecs (Maybe (ServerSettings -> ServerSettings))]
+    settings :: [SectionsSpec (Maybe (ServerSettings -> ServerSettings))]
     settings =
-      [ opt "name" ssName valuesSpec
+      [ opt "name" ssName anySpec
         "The name used to identify this server in the client"
 
       , req "hostname" ssHostName stringSpec
@@ -184,22 +205,22 @@ serverSpec = sectionsSpec "server-settings" $
       , req "nick" ssNicks nicksSpec
         "Nicknames to connect with in order"
 
-      , opt "password" ssPassword valuesSpec
+      , opt "password" ssPassword anySpec
         "Server password"
 
-      , req "username" ssUser valuesSpec
+      , req "username" ssUser anySpec
         "Second component of _!_@_ usermask"
 
-      , req "realname" ssReal valuesSpec
+      , req "realname" ssReal anySpec
         "\"GECOS\" name sent to server visible in /whois"
 
-      , req "userinfo" ssUserInfo valuesSpec
+      , req "userinfo" ssUserInfo anySpec
         "CTCP userinfo (currently unused)"
 
-      , opt "sasl-username" ssSaslUsername valuesSpec
+      , opt "sasl-username" ssSaslUsername anySpec
         "Username for SASL authentication to NickServ"
 
-      , opt "sasl-password" ssSaslPassword valuesSpec
+      , opt "sasl-password" ssSaslPassword anySpec
         "Password for SASL authentication to NickServ"
 
       , opt "sasl-ecdsa-key" ssSaslEcdsaFile stringSpec
@@ -232,16 +253,16 @@ serverSpec = sectionsSpec "server-settings" $
       , req "chanserv-channels" ssChanservChannels (listSpec identifierSpec)
         "Channels with ChanServ permissions available"
 
-      , req "flood-penalty" ssFloodPenalty valuesSpec
+      , req "flood-penalty" ssFloodPenalty anySpec
         "RFC 1459 rate limiting, seconds of penalty per message (default 2)"
 
-      , req "flood-threshold" ssFloodThreshold valuesSpec
+      , req "flood-threshold" ssFloodThreshold anySpec
         "RFC 1459 rate limiting, seconds of allowed penalty accumulation (default 10)"
 
-      , req "message-hooks" ssMessageHooks valuesSpec
+      , req "message-hooks" ssMessageHooks anySpec
         "Special message hooks to enable: \"buffextras\" available"
 
-      , req "reconnect-attempts" ssReconnectAttempts valuesSpec
+      , req "reconnect-attempts" ssReconnectAttempts anySpec
         "Number of reconnection attempts on lost connection"
 
       , req "autoconnect" ssAutoconnect yesOrNoSpec
@@ -255,32 +276,70 @@ serverSpec = sectionsSpec "server-settings" $
 
       , opt "protocol-family" ssProtocolFamily protocolFamilySpec
         "IP protocol family to use for this connection"
+
+      , req "sts" ssSts yesOrNoSpec
+        "Honor server STS policies forcing TLS connections"
+
+      , opt "tls-cert-fingerprint" ssTlsCertFingerprint fingerprintSpec
+        "Check SHA1, SHA256, or SHA512 certificate fingerprint"
+
+      , opt "tls-pubkey-fingerprint" ssTlsPubkeyFingerprint fingerprintSpec
+        "Check SHA1, SHA256, or SHA512 public key fingerprint"
       ]
 
+-- | Match fingerprints in plain hex or colon-delimited bytes.
+-- SHA-1 is 20 bytes. SHA-2-256 is 32 bytes. SHA-2-512 is 64 bytes.
+--
+-- @
+-- 00112233aaFF
+-- 00:11:22:33:aa:FF
+-- @
+fingerprintSpec :: ValueSpec Fingerprint
+fingerprintSpec =
+  customSpec "fingerprint" stringSpec $ \str ->
+    do bytes <- B.pack <$> traverse readWord8 (byteStrs str)
+       case B.length bytes of
+         20 -> Right (FingerprintSha1   bytes)
+         32 -> Right (FingerprintSha256 bytes)
+         64 -> Right (FingerprintSha512 bytes)
+         _  -> Left "expected 20, 32, or 64 bytes"
+  where
+    -- read a single byte in hex
+    readWord8 i =
+      case readHex i of
+        [(x,"")]
+          | 0 <= x, x < 256 -> Right (fromIntegral (x :: Integer))
+          | otherwise -> Left "byte out-of-bounds"
+        _ -> Left "bad hex-encoded byte"
+
+    byteStrs :: String -> [String]
+    byteStrs str
+      | ':' `elem` str = splitOn ":" str
+      | otherwise      = chunksOf 2  str
 
 -- | Specification for IP protocol family.
-protocolFamilySpec :: ValueSpecs Family
+protocolFamilySpec :: ValueSpec Family
 protocolFamilySpec =
       AF_INET   <$ atomSpec "inet"
   <!> AF_INET6  <$ atomSpec "inet6"
 
 
-nicksSpec :: ValueSpecs (NonEmpty Text)
-nicksSpec = oneOrNonemptySpec valuesSpec
+nicksSpec :: ValueSpec (NonEmpty Text)
+nicksSpec = oneOrNonemptySpec anySpec
 
 
-useTlsSpec :: ValueSpecs UseTls
+useTlsSpec :: ValueSpec UseTls
 useTlsSpec =
       UseTls         <$ atomSpec "yes"
   <!> UseInsecureTls <$ atomSpec "yes-insecure"
   <!> UseInsecure    <$ atomSpec "no"
 
 
-nickCompletionSpec :: ValueSpecs WordCompletionMode
+nickCompletionSpec :: ValueSpec WordCompletionMode
 nickCompletionSpec =
       defaultNickWordCompleteMode <$ atomSpec "default"
   <!> slackNickWordCompleteMode   <$ atomSpec "slack"
 
 
-identifierSpec :: ValueSpecs Identifier
-identifierSpec = mkId <$> valuesSpec
+identifierSpec :: ValueSpec Identifier
+identifierSpec = mkId <$> anySpec
