@@ -16,14 +16,14 @@ module Client.EventLoop
   , updateTerminalSize
   ) where
 
-import qualified Client.Authentication.Ecdsa as Ecdsa
-import           Client.CApi
+import           Client.CApi (popTimer)
 import           Client.Commands
-import           Client.Commands.Interpolation
-import           Client.Configuration (configJumpModifier, configKeyMap, configWindowNames)
+import           Client.Network.Async (recv)
+import           Client.Configuration (configJumpModifier, configKeyMap, configWindowNames, configDownloadDir)
 import           Client.Configuration.ServerSettings
 import           Client.EventLoop.Actions
 import           Client.EventLoop.Errors (exceptionToLines)
+import           Client.EventLoop.Network (clientResponse)
 import           Client.Hook
 import           Client.Hooks
 import           Client.Image
@@ -32,7 +32,9 @@ import           Client.Log
 import           Client.Message
 import           Client.Network.Async
 import           Client.State
+import           Client.State.DCC
 import qualified Client.State.EditBox     as Edit
+import           Client.State.Extensions
 import           Client.State.Focus
 import           Client.State.Network
 import           Control.Concurrent.STM
@@ -41,30 +43,32 @@ import           Control.Lens
 import           Control.Monad
 import           Data.ByteString (ByteString)
 import           Data.Foldable
+import           Data.Traversable
 import           Data.List
+import           Data.List.NonEmpty (NonEmpty, nonEmpty)
 import           Data.Maybe
-import           Data.Monoid
 import           Data.Ord
 import           Data.Text (Text)
+import qualified Data.HashMap.Strict as HashMap
 import qualified Data.Text as Text
 import qualified Data.Text.Encoding as Text
 import qualified Data.Text.Encoding.Error as Text
 import           Data.Time
 import           GHC.IO.Exception (IOErrorType(..), ioe_type)
 import           Graphics.Vty
-import           Irc.Codes
-import           Irc.Commands
 import           Irc.Message
 import           Irc.RawIrcMsg
 import           LensUtils
-import           Hookup
+import           Hookup (ConnectionFailure(..))
 
 
--- | Sum of the three possible event types the event loop handles
+-- | Sum of the five possible event types the event loop handles
 data ClientEvent
-  = VtyEvent Event -- ^ Key presses and resizing
-  | NetworkEvent NetworkEvent -- ^ Incoming network events
-  | TimerEvent NetworkId TimedAction -- ^ Timed action and the applicable network
+  = VtyEvent Event                        -- ^ Key presses and resizing
+  | NetworkEvents (NonEmpty (Text, NetworkEvent)) -- ^ Incoming network events
+  | TimerEvent Text TimedAction      -- ^ Timed action and the applicable network
+  | ExtTimerEvent Int                     -- ^ extension ID
+  | DCCUpdate DCCUpdate                   -- ^ Event on any transfer
 
 
 -- | Block waiting for the next 'ClientEvent'. This function will compute
@@ -75,31 +79,53 @@ getEvent ::
   IO ClientEvent
 getEvent vty st =
   do timer <- prepareTimer
-     atomically $
-       asum [ timer
-            , VtyEvent     <$> readTChan vtyEventChannel
-            , NetworkEvent <$> readTQueue (view clientEvents st)
-            ]
+     atomically (asum [timer, vtyEvent, networkEvents, dccUpdate])
   where
-    vtyEventChannel = _eventChannel (inputIface vty)
+    vtyEvent = VtyEvent <$> readTChan (_eventChannel (inputIface vty))
+
+    networkEvents =
+      do xs <- for (HashMap.toList (view clientConnections st)) $ \(network, conn) ->
+           do ys <- recv (view csSocket conn)
+              return (map ((,) network) ys)
+         case nonEmpty (concat xs) of
+           Just events1 -> return (NetworkEvents events1)
+           Nothing      -> retry
 
     prepareTimer =
       case earliestEvent st of
         Nothing -> return retry
-        Just (networkId,(runAt,action)) ->
+        Just (runAt,event) ->
           do now <- getCurrentTime
              let microsecs = truncate (1000000 * diffUTCTime runAt now)
              var <- registerDelay (max 0 microsecs)
              return $ do ready <- readTVar var
                          unless ready retry
-                         return (TimerEvent networkId action)
+                         return event
+
+    dccUpdate = DCCUpdate <$> readTChan (view clientDCCUpdates st)
 
 -- | Compute the earliest scheduled timed action for the client
-earliestEvent :: ClientState -> Maybe (NetworkId, (UTCTime, TimedAction))
-earliestEvent =
-  minimumByOf
-    (clientConnections . (ifolded <. folding nextTimedAction) . withIndex)
-    (comparing (fst . snd))
+earliestEvent :: ClientState -> Maybe (UTCTime, ClientEvent)
+earliestEvent st = earliest2 networkEvent extensionEvent
+  where
+    earliest2 (Just (time1, action1)) (Just (time2, action2))
+      | time1 < time2 = Just (time1, action1)
+      | otherwise     = Just (time2, action2)
+    earliest2 x y = mplus x y
+
+    mkEventN (network, (time, action)) = (time, TimerEvent network action)
+    networkEvent =
+      minimumByOf
+        (clientConnections . (ifolded <. folding nextTimedAction) . withIndex . to mkEventN)
+        (comparing fst)
+        st
+
+    mkEventE (i, (time,_,_,_,_)) = (time, ExtTimerEvent i)
+    extensionEvent =
+      minimumByOf
+        (clientExtensions . esActive . (ifolded <. folding popTimer) . withIndex . to mkEventE)
+        (comparing fst)
+        st
 
 -- | Apply this function to an initial 'ClientState' to launch the client.
 eventLoop :: Vty -> ClientState -> IO ()
@@ -112,15 +138,25 @@ eventLoop vty st =
 
      event <- getEvent vty st'
      case event of
-       TimerEvent networkId action  -> eventLoop vty =<< doTimerEvent networkId action st'
-       VtyEvent vtyEvent -> traverse_ (eventLoop vty) =<< doVtyEvent vty vtyEvent st'
-       NetworkEvent networkEvent ->
-         eventLoop vty =<<
-         case networkEvent of
-           NetworkLine  net time line -> doNetworkLine  net time line st'
-           NetworkError net time ex   -> doNetworkError net time ex st'
-           NetworkOpen  net time      -> doNetworkOpen  net time st'
-           NetworkClose net time      -> doNetworkClose net time st'
+       ExtTimerEvent i ->
+         eventLoop vty =<< clientExtTimer i st'
+       TimerEvent networkId action ->
+         eventLoop vty =<< doTimerEvent networkId action st'
+       VtyEvent vtyEvent ->
+         traverse_ (eventLoop vty) =<< doVtyEvent vty vtyEvent st'
+       NetworkEvents networkEvents ->
+         eventLoop vty =<< foldM doNetworkEvent st' networkEvents
+       DCCUpdate upd ->
+         eventLoop vty =<< doDCCUpdate upd st'
+
+-- | Apply a single network event to the client state.
+doNetworkEvent :: ClientState -> (Text, NetworkEvent) -> IO ClientState
+doNetworkEvent st (net, networkEvent) =
+  case networkEvent of
+    NetworkLine  time line -> doNetworkLine  net time line st
+    NetworkError time ex   -> doNetworkError net time ex st
+    NetworkOpen  time msg  -> doNetworkOpen  net time msg st
+    NetworkClose time      -> doNetworkClose net time st
 
 -- | Sound the terminal bell assuming that the @BEL@ control code
 -- is supported.
@@ -133,11 +169,12 @@ processLogEntries =
 
 -- | Respond to a network connection successfully connecting.
 doNetworkOpen ::
-  NetworkId   {- ^ network id   -} ->
+  Text        {- ^ network name -} ->
   ZonedTime   {- ^ event time   -} ->
+  [Text]      {- ^ rendered certificate -} ->
   ClientState {- ^ client state -} ->
   IO ClientState
-doNetworkOpen networkId time st =
+doNetworkOpen networkId time cert st =
   case view (clientConnections . at networkId) st of
     Nothing -> error "doNetworkOpen: Network missing"
     Just cs ->
@@ -146,14 +183,14 @@ doNetworkOpen networkId time st =
                      , _msgNetwork = view csNetwork cs
                      , _msgBody    = NormalBody "connection opened"
                      }
+         let cs' = cs & csLastReceived .~ (Just $! zonedTimeToUTC time)
+                      & csCertificate  .~ cert
          return $! recordNetworkMessage msg
-                 $ overStrict (clientConnections . ix networkId . csLastReceived)
-                              (\old -> old `seq` Just $! zonedTimeToUTC time)
-                              st
+                 $ setStrict (clientConnections . ix networkId) cs' st
 
 -- | Respond to a network connection closing normally.
 doNetworkClose ::
-  NetworkId   {- ^ network id   -} ->
+  Text        {- ^ network name -} ->
   ZonedTime   {- ^ event time   -} ->
   ClientState {- ^ client state -} ->
   IO ClientState
@@ -169,14 +206,14 @@ doNetworkClose networkId time st =
 
 -- | Respond to a network connection closing abnormally.
 doNetworkError ::
-  NetworkId     {- ^ failed network     -} ->
+  Text          {- ^ failed network     -} ->
   ZonedTime     {- ^ current time       -} ->
   SomeException {- ^ termination reason -} ->
   ClientState   {- ^ client state       -} ->
   IO ClientState
 doNetworkError networkId time ex st =
   do let (cs,st1) = removeNetwork networkId st
-         st2 = foldl' (\acc msg -> recordError time cs (Text.pack msg) acc) st1
+         st2 = foldl' (\acc msg -> recordError time (view csNetwork cs) (Text.pack msg) acc) st1
              $ exceptionToLines ex
      reconnectLogic ex cs st2
 
@@ -189,7 +226,7 @@ reconnectLogic ex cs st
 
   | shouldReconnect =
       do (attempts, mbDisconnectTime) <- computeRetryInfo
-         addConnection attempts mbDisconnectTime (view csNetwork cs) st
+         addConnection attempts mbDisconnectTime Nothing (view csNetwork cs) st
 
   | otherwise = return st
 
@@ -217,7 +254,7 @@ reconnectLogic ex cs st
 -- | Respond to an IRC protocol line. This will parse the message, updated the
 -- relevant connection state and update the UI buffers.
 doNetworkLine ::
-  NetworkId   {- ^ Network ID of message            -} ->
+  Text        {- ^ Network name                     -} ->
   ZonedTime   {- ^ current time                     -} ->
   ByteString  {- ^ Raw IRC message without newlines -} ->
   ClientState {- ^ client state                     -} ->
@@ -230,13 +267,10 @@ doNetworkLine networkId time line st =
       case parseRawIrcMsg (asUtf8 line) of
         Nothing ->
           do let msg = Text.pack ("Malformed message: " ++ show line)
-             return $! recordError time cs msg st
+             return $! recordError time (view csNetwork cs) msg st
 
         Just raw ->
-          do (st1,passed) <- clientPark st $ \ptr ->
-                               notifyExtensions ptr network raw
-                                 (view (clientExtensions . esActive) st)
-
+          do (st1,passed) <- clientNotifyExtensions network raw st
 
              if not passed then return st1 else do
 
@@ -267,97 +301,14 @@ doNetworkLine networkId time line st =
                                         , _msgBody    = IrcBody irc'
                                         }
 
-                    let (replies, st3) = applyMessageToClientState time irc networkId cs st2
+                    let (replies, dccUp, st3) =
+                          applyMessageToClientState time irc networkId cs st2
 
+                    traverse_ (atomically . writeTChan (view clientDCCUpdates st3)) dccUp
                     traverse_ (sendMsg cs) replies
                     clientResponse time' irc cs st3
 
 
--- | Client-level responses to specific IRC messages.
--- This is in contrast to the connection state tracking logic in
--- "Client.NetworkState"
-clientResponse :: ZonedTime -> IrcMsg -> NetworkState -> ClientState -> IO ClientState
-clientResponse now irc cs st =
-  case irc of
-    Reply RPL_WELCOME _ ->
-      -- run connection commands with the network focused and restore it afterward
-      do let focus = NetworkFocus (view csNetwork cs)
-         st' <- foldM (processConnectCmd now cs)
-                      (set clientFocus focus st)
-                      (view (csSettings . ssConnectCmds) cs)
-         return $! set clientFocus (view clientFocus st) st'
-
-    Authenticate challenge
-      | AS_EcdsaWaitChallenge <- view csAuthenticationState cs ->
-         processSaslEcdsa now challenge cs st
-
-    _ -> return st
-
-
-processSaslEcdsa ::
-  ZonedTime    {- ^ message time  -} ->
-  Text         {- ^ challenge     -} ->
-  NetworkState {- ^ network state -} ->
-  ClientState  {- ^ client state  -} ->
-  IO ClientState
-processSaslEcdsa now challenge cs st =
-  case view ssSaslEcdsaFile ss of
-    Nothing ->
-      do sendMsg cs ircCapEnd
-         return $! recordError now cs "panic: ecdsatool malformed output" st
-
-    Just path ->
-      do res <- Ecdsa.computeResponse path challenge
-         case res of
-           Left e ->
-             do sendMsg cs ircCapEnd
-                return $! recordError now cs (Text.pack e) st
-           Right resp ->
-             do sendMsg cs (ircAuthenticate resp)
-                return $! set asLens AS_None st
-  where
-    ss = view csSettings cs
-    asLens = clientConnections . ix (view csNetworkId cs) . csAuthenticationState
-
-
-processConnectCmd ::
-  ZonedTime       {- ^ now             -} ->
-  NetworkState    {- ^ current network -} ->
-  ClientState     {- ^ client state    -} ->
-  [ExpansionChunk]{- ^ command         -} ->
-  IO ClientState
-processConnectCmd now cs st0 cmdTxt =
-  do dc <- forM disco $ \t ->
-             Text.pack . formatTime defaultTimeLocale "%H:%M:%S"
-               <$> utcToLocalZonedTime t
-     let failureCase e = recordError now cs ("Bad connect-cmd: " <> e)
-     case resolveMacroExpansions (commandExpansion dc st0) (const Nothing) cmdTxt of
-       Nothing -> return $! failureCase "Unable to expand connect command" st0
-       Just cmdTxt' ->
-         do res <- executeUserCommand dc (Text.unpack cmdTxt') st0
-            return $! case res of
-              CommandFailure st -> failureCase cmdTxt' st
-              CommandSuccess st -> st
-              CommandQuit    st -> st -- not supported
- where
- disco = case view csPingStatus cs of
-   PingConnecting _ tm -> tm
-   _ -> Nothing
-
-
-recordError ::
-  ZonedTime       {- ^ now             -} ->
-  NetworkState    {- ^ current network -} ->
-  Text            {- ^ error message   -} ->
-  ClientState     {- ^ client state    -} ->
-  ClientState
-recordError now cs msg =
-  recordNetworkMessage ClientMessage
-    { _msgTime    = now
-    , _msgNetwork = view csNetwork cs
-    , _msgBody    = ErrorBody msg
-    }
-
 -- | Find the ZNC provided server time
 computeEffectiveTime :: ZonedTime -> [TagEntry] -> ZonedTime
 computeEffectiveTime time tags = fromMaybe time zncTime
@@ -419,8 +370,11 @@ doAction ::
 doAction vty action st =
 
   let continue !out -- detect when chains of M-a are broken
-        | action == ActJumpToActivity = return (Just out)
-        | otherwise = return $! Just $! set clientActivityReturn (view clientFocus out) out
+        | action == ActJumpToActivity =
+            let upd Nothing = Just $! view clientFocus st
+                upd x       = x
+            in return $! Just $! over clientActivityReturn upd out
+        | otherwise = return (Just out)
 
       changeEditor  f = continue (over clientTextBox f st)
       changeContent f = changeEditor
@@ -428,7 +382,7 @@ doAction vty action st =
                       . set  Edit.lastOperation Edit.OtherOperation
 
       mbChangeEditor f =
-        case clientTextBox f st of
+        case traverseOf clientTextBox f st of
           Nothing -> continue $! set clientBell True st
           Just st' -> continue st'
   in
@@ -504,6 +458,12 @@ doCommandResult clearOnSuccess res =
     CommandFailure st -> continue (set clientBell True st)
 
 
+-- | Actions to be run when exiting the client.
+clientShutdown :: ClientState -> IO ()
+clientShutdown st = () <$ clientStopExtensions st
+ -- other shutdown stuff might be added here later
+
+
 -- | Execute the the command on the first line of the text box
 executeInput ::
   ClientState {- ^ client state -} ->
@@ -513,11 +473,32 @@ executeInput st = execute (clientFirstLine st) st
 
 -- | Respond to a timer event.
 doTimerEvent ::
-  NetworkId   {- ^ Network related to event -} ->
+  Text        {- ^ Network related to event -} ->
   TimedAction {- ^ Action to perform        -} ->
   ClientState {- ^ client state             -} ->
   IO ClientState
 doTimerEvent networkId action =
   traverseOf
-    (clientConnections . ix networkId)
+    (clientConnection networkId)
     (applyTimedAction action)
+
+doDCCUpdate :: DCCUpdate -> ClientState -> IO ClientState
+doDCCUpdate upd st0 =
+  case upd of
+    PercentUpdate k newVal -> return $ set (commonLens k . dtProgress) newVal st0
+    SocketInterrupted k    -> reportKill k LostConnection
+    UserInterrupted k      -> reportKill k UserKilled
+    Finished k             -> reportKill k CorrectlyFinished
+    accept@(Accept k _ _) ->
+      do let dccState0 = view clientDCC st0
+             dccState1 = acceptUpdate accept dccState0
+             -- st1 = set clientDCC dccState1 st0
+
+             updChan = view clientDCCUpdates st0
+             mdir = view (clientConfig . configDownloadDir) st0
+
+         dccState2 <- supervisedDownload mdir k updChan dccState1
+         return $ set clientDCC dccState2 st0
+  where
+    reportKill k status = return $ over clientDCC (reportStopWithStatus k status) st0
+    commonLens k = clientDCC . dsTransfers . at k . _Just
