@@ -34,7 +34,6 @@ import           Data.Map (Map)
 import qualified Data.Map as Map
 import           Data.HashMap.Lazy (HashMap)
 import qualified Data.HashMap.Lazy as HashMap
-import           Data.Semigroup ((<>))
 import           Data.Text (Text)
 import qualified Data.Text as Text
 import           Text.Read
@@ -101,9 +100,11 @@ keyMapEntries (KeyMap m) =
      , (k,a) <- Map.toList m1
      ]
 
-instance Spec Action where
-  valuesSpec = customSpec "action" anyAtomSpec $ \a ->
-                fst <$> HashMap.lookup a actionInfos
+instance HasSpec Action where
+  anySpec = customSpec "action" anyAtomSpec
+          $ \a -> case HashMap.lookup a actionInfos of
+                    Nothing -> Left "unknown action"
+                    Just x  -> Right (fst x)
 
 
 -- | Names and default key bindings for each action.
