@@ -25,9 +25,10 @@ import           Client.State.Focus
 import           Client.State.Network
 import           Client.State.Window
 import           Control.Lens
+import           Data.Foldable (for_)
 import qualified Data.Map.Strict as Map
 import           Data.Maybe
-import           Data.Semigroup
+import           Data.HashMap.Strict (HashMap)
 import           Data.Text (Text)
 import qualified Data.Text as Text
 import           Graphics.Vty.Attributes
@@ -106,42 +107,53 @@ scrollImage st
   | otherwise = infoBubble (string attr "scroll")
   where
     pal  = clientPalette st
-    attr = view palLabel pal
+    attr = view palError pal
 
 
 -- | Indicate when the client is potentially showing a subset of the
 -- available chat messages.
 filterImage :: ClientState -> Image'
 filterImage st =
-  case clientActiveRegex st of
+  case clientMatcher st of
     Nothing -> mempty
     Just {} -> infoBubble (string attr "filtered")
   where
     pal  = clientPalette st
-    attr = view palLabel pal
+    attr = view palError pal
 
 
 -- | Indicate the current connection health. This will either indicate
 -- that the connection is being established or that a ping has been
 -- sent or long the previous ping round-trip was.
 latencyImage :: ClientState -> Image'
-latencyImage st =
-  case views clientFocus focusNetwork st of
-    Nothing      -> mempty
-    Just network ->
-      case preview (clientConnection network) st of
-        Nothing -> infoBubble (string (view palError pal) "offline")
-        Just cs ->
-          case view csPingStatus cs of
-            PingNever          -> mempty
-            PingSent {}        -> latency "ping sent"
-            PingLatency delta  -> latency (showFFloat (Just 2) delta "s")
-            PingConnecting n _ ->
-              infoBubble (string (view palLatency pal) "connecting" <>
-                          retryImage n)
+latencyImage st = either id id $
+
+  do network <- -- no network -> no image
+       case views clientFocus focusNetwork st of
+         Nothing  -> Left mempty
+         Just net -> Right net
+
+     cs <- -- detect when offline
+       case preview (clientConnection network) st of
+         Nothing -> Left (infoBubble (string (view palError pal) "offline"))
+         Just cs -> Right cs
+
+     -- render latency if one is stored
+     for_ (view csLatency cs) $ \latency ->
+       Left (latencyBubble (showFFloat (Just 2) (realToFrac latency :: Double) "s"))
+
+     Right $ case view csPingStatus cs of
+
+       PingSent {} -> latencyBubble "wait"
+
+       PingConnecting n _ ->
+         infoBubble (string (view palLatency pal) "connecting" <> retryImage n)
+
+       PingNone -> mempty -- just connected no ping sent yet
+
   where
-    pal     = clientPalette st
-    latency = infoBubble . string (view palLatency pal)
+    pal           = clientPalette st
+    latencyBubble = infoBubble . string (view palLatency pal)
 
     retryImage n
       | n > 0     = ": " <> string (view palLabel pal) ("retry " ++ show n)
@@ -272,14 +284,27 @@ myNickImage st =
         Nothing -> Vty.emptyImage
         Just cs -> Vty.string (view palSigil pal) myChanModes
            Vty.<|> Vty.text' defAttr (idText nick)
-           Vty.<|> parens defAttr (Vty.string defAttr ('+' : view csModes cs))
+           Vty.<|> parens defAttr
+                     (unpackImage $
+                      modesImage (view palUModes pal) (view csModes cs) <>
+                      snomaskImage)
           where
-            nick      = view csNick cs
+            nick = view csNick cs
+
+            snomaskImage
+              | null (view csSnomask cs) = ""
+              | otherwise                = " " <> modesImage (view palSnomask pal) (view csSnomask cs)
+
             myChanModes =
               case mbChan of
                 Nothing   -> []
                 Just chan -> view (csChannels . ix chan . chanUsers . ix nick) cs
 
+modesImage :: HashMap Char Attr -> String -> Image'
+modesImage pal modes = "+" <> foldMap modeImage modes
+  where
+    modeImage m =
+      char (fromMaybe defAttr (view (at m) pal)) m
 
 subfocusImage :: ClientState -> Image'
 subfocusImage st = foldMap infoBubble (viewSubfocusLabel pal subfocus)
@@ -322,29 +347,29 @@ viewFocusLabel st focus =
 channelModesImage :: Text -> Identifier -> ClientState -> Image'
 channelModesImage network channel st =
   case preview (clientConnection network . csChannels . ix channel . chanModes) st of
-    Just modeMap | not (null modeMap) ->
-        string defAttr (" +" ++ modes) <>
-        mconcat [ char defAttr ' ' <> text' defAttr arg | arg <- args, not (Text.null arg) ]
-      where (modes,args) = unzip (Map.toList modeMap)
-    _ -> mempty
+    Just modeMap | not (null modeMap) -> " " <> modesImage (view palCModes pal) (Map.keys modeMap)
+    _                                 -> mempty
+  where
+    pal = clientPalette st
 
 viewSubfocusLabel :: Palette -> Subfocus -> Maybe Image'
 viewSubfocusLabel pal subfocus =
   case subfocus of
-    FocusMessages -> Nothing
+    FocusMessages     -> Nothing
     FocusWindows filt -> Just $ string (view palLabel pal) "windows" <>
                                 opt (windowFilterName filt)
-    FocusInfo     -> Just $ string (view palLabel pal) "info"
-    FocusUsers    -> Just $ string (view palLabel pal) "users"
-    FocusMentions -> Just $ string (view palLabel pal) "mentions"
-    FocusPalette  -> Just $ string (view palLabel pal) "palette"
-    FocusDigraphs -> Just $ string (view palLabel pal) "digraphs"
-    FocusKeyMap   -> Just $ string (view palLabel pal) "keymap"
-    FocusHelp mb  -> Just $ string (view palLabel pal) "help" <>
-                            opt mb
-    FocusIgnoreList -> Just $ string (view palLabel pal) "ignores"
-    FocusRtsStats -> Just $ string (view palLabel pal) "rtsstats"
-    FocusMasks m  -> Just $ mconcat
+    FocusInfo         -> Just $ string (view palLabel pal) "info"
+    FocusUsers        -> Just $ string (view palLabel pal) "users"
+    FocusMentions     -> Just $ string (view palLabel pal) "mentions"
+    FocusDCC          -> Just $ string (view palLabel pal) "dcc"
+    FocusPalette      -> Just $ string (view palLabel pal) "palette"
+    FocusDigraphs     -> Just $ string (view palLabel pal) "digraphs"
+    FocusKeyMap       -> Just $ string (view palLabel pal) "keymap"
+    FocusHelp mb      -> Just $ string (view palLabel pal) "help" <> opt mb
+    FocusIgnoreList   -> Just $ string (view palLabel pal) "ignores"
+    FocusRtsStats     -> Just $ string (view palLabel pal) "rtsstats"
+    FocusCert{}       -> Just $ string (view palLabel pal) "cert"
+    FocusMasks m      -> Just $ mconcat
       [ string (view palLabel pal) "masks"
       , char defAttr ':'
       , char (view palLabel pal) m
