@@ -29,11 +29,9 @@ import           Client.State
 import           Client.State.Focus
 import qualified Client.State.EditBox as Edit
 import           Control.Lens
-import           Data.Char
 import qualified Data.HashSet as HashSet
 import           Data.HashSet (HashSet)
 import           Data.List
-import           Data.Semigroup
 import qualified Data.Text as Text
 import           Graphics.Vty.Attributes
 import qualified Graphics.Vty.Image as Vty
@@ -48,11 +46,12 @@ textboxImage width st
   where
   macros = views (clientConfig . configMacros) (fmap macroSpec) st
   (txt, content) =
-     views (clientTextBox . Edit.content) (renderContent st myNick nicks macros pal) st
+     renderContent st myNick nicks macros pal
+       (view (clientTextBox . Edit.content) st)
 
   lineImage = unpackImage (beginning <> content <> ending)
 
-  leftOfCurWidth = myWcswidth ('^':txt)
+  leftOfCurWidth = 1 + txt
 
   croppedImage = Vty.resizeWidth width
                $ Vty.cropLeft (Vty.imageWidth lineImage - newOffset) lineImage
@@ -97,8 +96,8 @@ renderContent ::
   Recognizer MacroSpec {- ^ macro completions                     -} ->
   Palette              {- ^ palette                               -} ->
   Edit.Content         {- ^ content                               -} ->
-  (String, Image')     {- ^ plain text rendering, image rendering -}
-renderContent st myNick nicks macros pal c = (txt, wholeImg)
+  (Int, Image')        {- ^ left-of-cursor width, image rendering -}
+renderContent st myNick nicks macros pal c = (leftLen, wholeImg)
   where
   as  = reverse (view Edit.above c)
   bs  = view Edit.below c
@@ -108,30 +107,19 @@ renderContent st myNick nicks macros pal c = (txt, wholeImg)
   leftCur = take (view Edit.pos cur) (view Edit.text cur)
 
   -- ["one","two"] "three" --> "two one three"
-  txt = foldl (\acc x -> x ++ ' ' : acc) leftCur as
+  leftLen = length leftCur + length leftImgs + sum (map imageWidth leftImgs)
 
   rndr = renderLine st pal myNick nicks macros
 
+  leftImgs = map (rndr False) as
+
   wholeImg = mconcat
            $ intersperse (plainText "\n")
-           $ map (rndr False) as
+           $ leftImgs
           ++ rndr True curTxt
            : map (rndr False) bs
 
 
--- | Version of 'wcwidth' that accounts for how control characters are
--- rendered
-myWcwidth :: Char -> Int
-myWcwidth x
-  | isControl x = 1
-  | otherwise   = Vty.wcwidth x
-
--- | Version of 'wcswidth' that accounts for how control characters are
--- rendered
-myWcswidth :: String -> Int
-myWcswidth = sum . map myWcwidth
-
-
 -- | Render the active text box line using command highlighting and
 -- placeholders, and WYSIWYG mIRC formatting control characters.
 renderLine ::
@@ -143,35 +131,37 @@ renderLine ::
   Bool                 {- ^ focused      -} ->
   String               {- ^ input text   -} ->
   Image'               {- ^ output image -}
-renderLine st pal myNick nicks macros focused ('/':xs) =
-  char defAttr '/' <> string attr cleanCmd <> continue rest
-  where
-    specAttr spec =
-      case parse st spec rest of
-        Nothing -> view palCommand      pal
-        Just{}  -> view palCommandReady pal
+renderLine st pal myNick nicks macros focused input =
 
-    (cmd, rest) = break isSpace xs
-    cleanCmd = map cleanChar cmd
+  case span (' '==) input of
+    (spcs, '/':xs) -> string defAttr spcs <> char defAttr '/'
+                   <> string attr cleanCmd <> continue rest
+      where
+        specAttr spec =
+          case parse st spec rest of
+            Nothing -> view palCommand      pal
+            Just{}  -> view palCommandReady pal
 
-    allCommands = (Left <$> macros) <> (Right <$> commands)
-    (attr, continue)
-      = case recognize (Text.pack cmd) allCommands of
-          Exact (Right Command{cmdArgumentSpec = spec}) ->
-            ( specAttr spec
-            , render pal st focused spec
-            )
-          Exact (Left (MacroSpec spec)) ->
-            ( specAttr spec
-            , render pal st focused spec
-            )
-          Prefix _ ->
-            ( view palCommandPrefix pal
-            , parseIrcTextWithNicks pal myNick nicks focused . Text.pack
-            )
-          Invalid ->
-            ( view palCommandError pal
-            , parseIrcTextWithNicks pal myNick nicks focused . Text.pack
-            )
+        (cmd, rest) = break (' '==) xs
+        cleanCmd = map cleanChar cmd
 
-renderLine _ pal myNick nicks _ focused xs = parseIrcTextWithNicks pal myNick nicks focused (Text.pack xs)
+        allCommands = (Left <$> macros) <> (Right <$> commands)
+        (attr, continue)
+          = case recognize (Text.pack cmd) allCommands of
+              Exact (Right Command{cmdArgumentSpec = spec}) ->
+                ( specAttr spec
+                , render pal st focused spec
+                )
+              Exact (Left (MacroSpec spec)) ->
+                ( specAttr spec
+                , render pal st focused spec
+                )
+              Prefix _ ->
+                ( view palCommandPrefix pal
+                , parseIrcTextWithNicks pal myNick nicks focused . Text.pack
+                )
+              Invalid ->
+                ( view palCommandError pal
+                , parseIrcTextWithNicks pal myNick nicks focused . Text.pack
+                )
+    _ -> parseIrcTextWithNicks pal myNick nicks focused (Text.pack input)
