@@ -39,6 +39,7 @@ import           Irc.Message
 import           Irc.Identifier
 import           Irc.UserInfo
 import           Irc.Codes
+import           Client.State.DCC (isSend)
 
 data MessageBody
   = IrcBody    !IrcMsg
@@ -63,6 +64,9 @@ data IrcSummary
   | ReplySummary {-# UNPACK #-} !ReplyCode
   | ChatSummary {-# UNPACK #-} !UserInfo
   | CtcpSummary {-# UNPACK #-} !Identifier
+  | DccSendSummary {-# UNPACK #-} !Identifier
+  | ChngSummary {-# UNPACK #-} !Identifier -- ^ Chghost command
+  | AcctSummary {-# UNPACK #-} !Identifier -- ^ Account command
   | NoSummary
   deriving (Eq, Show)
 
@@ -83,14 +87,17 @@ msgSummary (NormalBody _  ) = NoSummary
 ircSummary :: IrcMsg -> IrcSummary
 ircSummary msg =
   case msg of
-    Join who _      -> JoinSummary (userNick who)
+    Join who _ _    -> JoinSummary (userNick who)
     Part who _ _    -> PartSummary (userNick who)
     Quit who _      -> QuitSummary (userNick who)
     Nick who who'   -> NickSummary (userNick who) who'
     Privmsg who _ _ -> ChatSummary who
     Notice who _ _  -> ChatSummary who
     Ctcp who _ "ACTION" _ -> ChatSummary who
+    Ctcp who _ "DCC" txt | isSend txt -> DccSendSummary (userNick who)
     Ctcp who _ _ _ -> CtcpSummary (userNick who)
     CtcpNotice who _ _ _ -> ChatSummary who
     Reply code _    -> ReplySummary code
+    Account who _   -> AcctSummary (userNick who)
+    Chghost who _ _ -> ChngSummary (userNick who)
     _               -> NoSummary
