@@ -20,6 +20,8 @@ module Client.State
   , clientTextBox
   , clientTextBoxOffset
   , clientConnections
+  , clientDCC
+  , clientDCCUpdates
   , clientWidth
   , clientHeight
   , clientEvents
@@ -32,7 +34,6 @@ module Client.State
   , clientActivityBar
   , clientShowPing
   , clientSubfocus
-  , clientNetworkMap
   , clientIgnores
   , clientIgnoreMask
   , clientConnection
@@ -45,15 +46,11 @@ module Client.State
   , clientLayout
   , clientRtsStats
   , clientConfigPath
+  , clientStsPolicy
 
   -- * Client operations
   , withClientState
-  , clientStartExtensions
-  , clientShutdown
-  , clientPark
-  , clientMatcher
-  , clientMatcher'
-  , clientActiveRegex
+  , clientMatcher, Matcher(..), buildMatcher
   , clientToggleHideMeta
   , clientHighlightsNetwork
   , channelUserList
@@ -67,6 +64,7 @@ module Client.State
   , addConnection
   , removeNetwork
   , clientTick
+  , queueDCCTransfer
   , applyMessageToClientState
   , clientHighlights
   , clientWindowNames
@@ -80,6 +78,7 @@ module Client.State
   -- * Add messages to buffers
   , recordChannelMessage
   , recordNetworkMessage
+  , recordError
   , recordIrcMessage
 
   -- * Focus manipulation
@@ -99,6 +98,8 @@ module Client.State
   -- * Extensions
   , ExtensionState
   , esActive
+  , esMVar
+  , esStablePtr
 
   -- * URL view
   , urlPattern
@@ -110,6 +111,7 @@ import           Client.CApi
 import           Client.Commands.WordCompletion
 import           Client.Configuration
 import           Client.Configuration.ServerSettings
+import           Client.Configuration.Sts
 import           Client.Image.Message
 import           Client.Image.Palette
 import           Client.Log
@@ -121,6 +123,7 @@ import qualified Client.State.EditBox as Edit
 import           Client.State.Focus
 import           Client.State.Network
 import           Client.State.Window
+import           Client.State.DCC
 import           Control.Applicative
 import           Control.Concurrent.MVar
 import           Control.Concurrent.STM
@@ -128,7 +131,6 @@ import           Control.Exception
 import           Control.Lens
 import           Control.Monad
 import           Data.Foldable
-import           Data.Either
 import           Data.HashMap.Strict (HashMap)
 import qualified Data.HashMap.Strict as HashMap
 import           Data.HashSet (HashSet)
@@ -143,7 +145,6 @@ import           Data.Text (Text)
 import qualified Data.Text as Text
 import qualified Data.Text.Lazy as LText
 import           Data.Time
-import           Foreign.Ptr
 import           Foreign.StablePtr
 import           Irc.Codes
 import           Irc.Identifier
@@ -155,21 +156,23 @@ import           RtsStats (Stats)
 import           Text.Regex.TDFA
 import           Text.Regex.TDFA.String (compile)
 
+
 -- | All state information for the IRC client
 data ClientState = ClientState
   { _clientWindows           :: !(Map Focus Window) -- ^ client message buffers
   , _clientPrevFocus         :: !Focus              -- ^ previously focused buffer
-  , _clientActivityReturn    :: !Focus              -- ^ focus prior to jumping to activity
+  , _clientActivityReturn    :: !(Maybe Focus)      -- ^ focus prior to jumping to activity
   , _clientFocus             :: !Focus              -- ^ currently focused buffer
   , _clientSubfocus          :: !Subfocus           -- ^ current view mode
   , _clientExtraFocus        :: ![Focus]            -- ^ extra messages windows to view
 
-  , _clientConnections       :: !(IntMap NetworkState) -- ^ state of active connections
+  , _clientConnections       :: !(HashMap Text NetworkState) -- ^ state of active connections
   , _clientEvents            :: !(TQueue NetworkEvent)    -- ^ incoming network event queue
-  , _clientNetworkMap        :: !(HashMap Text NetworkId) -- ^ network name to connection ID
+  , _clientDCC               :: !DCCState                 -- ^ DCC subsystem
+  , _clientDCCUpdates        :: !(TChan DCCUpdate)        -- ^ DCC update events
 
   , _clientConfig            :: !Configuration            -- ^ client configuration
-  , _clientConfigPath        :: !(Maybe FilePath)         -- ^ client configuration file path
+  , _clientConfigPath        :: !FilePath                 -- ^ client configuration file path
 
   , _clientTextBox           :: !Edit.EditBox             -- ^ primary text box
   , _clientTextBoxOffset     :: !Int                      -- ^ size to crop from left of text box
@@ -180,7 +183,7 @@ data ClientState = ClientState
   , _clientDetailView        :: !Bool                     -- ^ use detailed rendering mode
   , _clientActivityBar       :: !Bool                     -- ^ visible activity bar
   , _clientShowPing          :: !Bool                     -- ^ visible ping time
-  , _clientRegex             :: Maybe Regex               -- ^ optional persistent filter
+  , _clientRegex             :: Maybe Matcher             -- ^ optional persistent filter
   , _clientLayout            :: !LayoutMode               -- ^ layout mode for split screen
 
   , _clientBell              :: !Bool                     -- ^ sound a bell next draw
@@ -192,44 +195,40 @@ data ClientState = ClientState
   , _clientLogQueue          :: ![LogLine]                -- ^ log lines ready to write
   , _clientErrorMsg          :: Maybe Text                -- ^ transient error box text
   , _clientRtsStats          :: Maybe Stats               -- ^ most recent GHC RTS stats
+
+  , _clientStsPolicy         :: !(HashMap Text StsPolicy) -- ^ STS policy entries
   }
 
+data Matcher = Matcher
+  { matcherBefore :: !Int
+  , matcherAfter  :: !Int
+  , matcherPred   :: LText.Text -> Bool
+  }
 
 -- | State of the extension API including loaded extensions and the mechanism used
 -- to support reentry into the Haskell runtime from the C API.
+--
+-- When executing inside an extension the mvar will contain the client state
+-- and the ID of the running extension.
 data ExtensionState = ExtensionState
-  { _esActive    :: [ActiveExtension]            -- ^ active extensions
-  , _esMVar      :: MVar ClientState             -- ^ 'MVar' used to with 'clientPark'
-  , _esStablePtr :: StablePtr (MVar ClientState) -- ^ 'StablePtr' used with 'clientPark'
+  { _esActive    :: IntMap ActiveExtension     -- ^ active extensions
+  , _esMVar      :: MVar ParkState             -- ^ 'MVar' used to with 'clientPark'
+  , _esStablePtr :: StablePtr (MVar ParkState) -- ^ 'StablePtr' used with 'clientPark'
   }
 
+-- | ID of active extension and stored client state
+type ParkState = (Int,ClientState)
+
 makeLenses ''ClientState
 makeLenses ''ExtensionState
 
-
--- | Prepare the client to support reentry from the extension API.
-clientPark ::
-  ClientState      {- ^ client state                                        -} ->
-  (Ptr () -> IO a) {- ^ continuation using the stable pointer to the client -} ->
-  IO (ClientState, a)
-clientPark st k =
-  do let mvar = view (clientExtensions . esMVar) st
-     putMVar mvar st
-     let token = views (clientExtensions . esStablePtr) castStablePtrToPtr st
-     res <- k token
-     st' <- takeMVar mvar
-     return (st', res)
-
 -- | 'Traversal' for finding the 'NetworkState' associated with a given network
 -- if that connection is currently active.
 clientConnection ::
   Applicative f =>
   Text {- ^ network -} ->
   LensLike' f ClientState NetworkState
-clientConnection network f st =
-  case view (clientNetworkMap . at network) st of
-    Nothing -> pure st
-    Just i  -> clientConnections (ix i f) st
+clientConnection network = clientConnections . ix network
 
 -- | The full top-most line that would be executed
 clientFirstLine :: ClientState -> String
@@ -240,26 +239,29 @@ clientLine :: ClientState -> (Int, String) {- ^ line number, line content -}
 clientLine = views (clientTextBox . Edit.line) (\(Edit.Line n t) -> (n, t))
 
 -- | Construct an initial 'ClientState' using default values.
-withClientState :: Maybe FilePath -> Configuration -> (ClientState -> IO a) -> IO a
+withClientState :: FilePath -> Configuration -> (ClientState -> IO a) -> IO a
 withClientState cfgPath cfg k =
 
   withExtensionState $ \exts ->
 
-  do events <- atomically newTQueue
+  do events    <- atomically newTQueue
+     dccEvents <- atomically newTChan
+     sts       <- readPolicyFile
      let ignoreIds = map mkId (view configIgnores cfg)
      k ClientState
         { _clientWindows           = _Empty # ()
-        , _clientNetworkMap        = _Empty # ()
         , _clientIgnores           = HashSet.fromList ignoreIds
         , _clientIgnoreMask        = buildMask ignoreIds
         , _clientConnections       = _Empty # ()
+        , _clientDCC               = emptyDCCState
+        , _clientDCCUpdates        = dccEvents
         , _clientTextBox           = Edit.defaultEditBox
         , _clientTextBoxOffset     = 0
         , _clientWidth             = 80
         , _clientHeight            = 25
         , _clientEvents            = events
         , _clientPrevFocus         = Unfocused
-        , _clientActivityReturn    = Unfocused
+        , _clientActivityReturn    = Nothing
         , _clientFocus             = Unfocused
         , _clientSubfocus          = FocusMessages
         , _clientExtraFocus        = []
@@ -276,6 +278,7 @@ withClientState cfgPath cfg k =
         , _clientLogQueue          = []
         , _clientErrorMsg          = Nothing
         , _clientRtsStats          = Nothing
+        , _clientStsPolicy         = sts
         }
 
 withExtensionState :: (ExtensionState -> IO a) -> IO a
@@ -283,7 +286,7 @@ withExtensionState k =
   do mvar <- newEmptyMVar
      bracket (newStablePtr mvar) freeStablePtr $ \stab ->
        k ExtensionState
-         { _esActive    = []
+         { _esActive    = IntMap.empty
          , _esMVar      = mvar
          , _esStablePtr = stab
          }
@@ -299,7 +302,7 @@ abortNetwork network st =
     Just cs -> do -- cancel the network thread
                   abortConnection ForcedDisconnect (view csSocket cs)
                   -- unassociate this network name from this network id
-                  return $! over clientNetworkMap (sans network) st
+                  return $! over clientConnections (sans network) st
 
 
 -- | Add a message to the window associated with a given channel
@@ -391,11 +394,17 @@ msgImportance msg st =
         Kick _ _ kicked _ | isMe kicked -> WLImportant
                           | otherwise   -> WLNormal
         Error{} -> WLImportant
+
+        -- channel information
+        Reply RPL_TOPIC _        -> WLBoring
+        Reply RPL_INVITING _     -> WLBoring
+
+        -- remaining replies go to network window
         Reply cmd _ ->
           case replyCodeType (replyCodeInfo cmd) of
             ErrorReply -> WLImportant
             _          -> WLNormal
-        _               -> WLBoring
+        _              -> WLBoring
 
 
 -- | Predicate for messages that should be ignored based on the
@@ -510,6 +519,18 @@ recordNetworkMessage msg st = updateTransientError focus msg
 
     cfg        = view clientConfig st
 
+recordError ::
+  ZonedTime       {- ^ now             -} ->
+  Text            {- ^ network         -} ->
+  Text            {- ^ error message   -} ->
+  ClientState     {- ^ client state    -} ->
+  ClientState
+recordError now net msg =
+  recordNetworkMessage ClientMessage
+    { _msgTime    = now
+    , _msgNetwork = net
+    , _msgBody    = ErrorBody msg
+    }
 
 -- | Record window line at the given focus creating the window if necessary
 recordWindowLine ::
@@ -611,41 +632,35 @@ channelUserList ::
 channelUserList network channel =
   views (clientConnection network . csChannels . ix channel . chanUsers) HashMap.keys
 
-
 -- | Returns the current filtering predicate if one is active.
 clientMatcher ::
-  ClientState          {- ^ client state       -} ->
-  Maybe (LText.Text -> Bool) {- ^ optional predicate -}
+  ClientState   {- ^ client state       -} ->
+  Maybe Matcher {- ^ optional predicate -}
 clientMatcher st =
-  do r <- clientActiveRegex st
-     return (matchTest r . LText.unpack)
-
--- | Strict version of 'clientMatcher'
-clientMatcher' ::
-  ClientState          {- ^ client state       -} ->
-  Maybe (Text -> Bool) {- ^ optional predicate -}
-clientMatcher' st =
-  do p <- clientMatcher st
-     return (p . LText.fromStrict)
-
-
--- | Construct a text matching predicate used to filter the message window.
-clientActiveRegex :: ClientState -> Maybe Regex
-clientActiveRegex st =
   case clientActiveCommand st of
-    Just ("grep" ,reStr) -> go True  reStr
-    Just ("grepi",reStr) -> go False reStr
+    Just ("grep" , reStr) -> buildMatcher reStr
     _ -> case view clientRegex st of
            Nothing -> Nothing
            Just r  -> Just r
+
+buildMatcher :: String -> Maybe Matcher
+buildMatcher = go (True, 0, 0)
   where
-    go sensitive reStr =
+    go (sensitive, b, a) reStr =
+      case dropWhile (' '==) reStr of
+        '-' : 'i' : ' ' : reStr'                                            -> go (False, b, a) reStr'
+        '-' : 'A' : reStr' | [(a' , ' ':reStr'')] <- reads reStr', a'  >= 0 -> go (sensitive, b, a') reStr''
+        '-' : 'B' : reStr' | [(b' , ' ':reStr'')] <- reads reStr', b'  >= 0 -> go (sensitive, b', a) reStr''
+        '-' : 'C' : reStr' | [(num, ' ':reStr'')] <- reads reStr', num >= 0 -> go (sensitive, num, num) reStr''
+        '-' : '-' : reStr' -> finish (sensitive, b, a) (drop 1 reStr')
+        _ -> finish (sensitive, b, a) reStr
+
+    finish (sensitive, b, a) reStr =
       case compile defaultCompOpt{caseSensitive=sensitive}
                    defaultExecOpt{captureGroups=False}
                    reStr of
         Left{}  -> Nothing
-        Right r -> Just r
-
+        Right r -> Just (Matcher b a (matchTest r . LText.unpack))
 
 -- | Compute the command and arguments currently in the textbox.
 clientActiveCommand ::
@@ -663,8 +678,9 @@ Right urlPattern =
   compile
     defaultCompOpt
     defaultExecOpt{captureGroups=False}
-    "https?://([[:alnum:]-]+\\.)*([[:alnum:]-]+)(:[[:digit:]]+)?(/[^[:cntrl:][:space:]]*)|\
-    \<https?://[^>]*>"
+    "https?://([[:alnum:]-]+\\.)*([[:alnum:]-]+)(:[[:digit:]]+)?(/[-0-9a-zA-Z$_.+!*'(),%?=:@/;~#]*)?|\
+    \<https?://[^>]*>|\
+    \\\(https?://[^\\)]*\\)"
 
 
 -- | Find all the URL matches using 'urlPattern' in a given 'Text' suitable
@@ -680,32 +696,27 @@ urlMatches txt = removeBrackets . extractText . (^?! ix 0)
     removeBrackets t =
       case Text.uncons t of
        Just ('<',t') | not (Text.null t') -> Text.init t'
+       Just ('(',t') | not (Text.null t') -> Text.init t'
        _                                  -> t
 
 -- | Remove a network connection and unlink it from the network map.
 -- This operation assumes that the network connection exists and should
 -- only be applied once per connection.
-removeNetwork :: NetworkId -> ClientState -> (NetworkState, ClientState)
-removeNetwork networkId st =
-  case (clientConnections . at networkId <<.~ Nothing) st of
+removeNetwork :: Text -> ClientState -> (NetworkState, ClientState)
+removeNetwork network st =
+  case (clientConnections . at network <<.~ Nothing) st of
     (Nothing, _  ) -> error "removeNetwork: network not found"
-    (Just cs, st1) ->
-      -- Only remove the network mapping if it hasn't already been replaced
-      -- with a new one. This can happen during reconnect in particular.
-      let network = view csNetwork cs in
-      forOf (clientNetworkMap . at network) st1 $ \mb ->
-        case mb of
-          Just i | i == networkId -> (cs,Nothing)
-          _                       -> (cs,mb)
+    (Just cs, st1) -> (cs, st1)
 
 -- | Start a new connection. The delay is used for reconnections.
 addConnection ::
   Int           {- ^ attempts                 -} ->
   Maybe UTCTime {- ^ optional disconnect time -} ->
+  Maybe Int     {- ^ STS upgrade port         -} ->
   Text          {- ^ network name             -} ->
   ClientState ->
   IO ClientState
-addConnection attempts lastTime network st =
+addConnection attempts lastTime stsUpgrade network st =
   do let defSettings = (view (clientConfig . configDefaults) st)
                      { _ssName = Just network
                      , _ssHostName = Text.unpack network
@@ -714,45 +725,65 @@ addConnection attempts lastTime network st =
          settings = fromMaybe defSettings
                   $ preview (clientConfig . configServers . ix network) st
 
-         i = nextAvailableKey (view clientConnections st)
+     now <- getCurrentTime
+     let stsUpgrade'
+           | Just{} <- stsUpgrade = stsUpgrade
+           | UseInsecure <- view ssTls settings
+           , let host = Text.pack (view ssHostName settings)
+           , Just policy <- view (clientStsPolicy . at host) st
+           , now < view stsExpiration policy
+           = Just (view stsPort policy)
+           | otherwise = Nothing
+
+         settings1 =
+           case stsUpgrade' of
+             Just port -> set ssPort (Just (fromIntegral port))
+                        $ set ssTls UseTls settings
+             Nothing   -> settings
+
 
          -- don't bother delaying on the first reconnect
          delay = 15 * max 0 (attempts - 1)
 
      c <- createConnection
             delay
-            i
-            settings
-            (view clientEvents st)
+            settings1
 
-     let cs = newNetworkState i network settings c (PingConnecting attempts lastTime)
+     let cs = newNetworkState network settings1 c (PingConnecting attempts lastTime)
      traverse_ (sendMsg cs) (initialMessages cs)
 
-     return $ set (clientNetworkMap . at network) (Just i)
-            $ set (clientConnections . at i) (Just cs) st
+     return $ set (clientConnections . at network) (Just cs) st
 
--- | Find the first unused key in the intmap starting at 0.
-nextAvailableKey :: IntMap a -> Int
-nextAvailableKey m = foldr aux id (IntMap.keys m) 0
-  where
-    aux k next i
-      | k == i    = next (i+1)
-      | otherwise = i
 
 applyMessageToClientState ::
   ZonedTime                  {- ^ timestamp                -} ->
   IrcMsg                     {- ^ message received         -} ->
-  NetworkId                  {- ^ message network          -} ->
+  Text                       {- ^ network name             -} ->
   NetworkState               {- ^ network connection state -} ->
   ClientState                {- ^ client state             -} ->
-  ([RawIrcMsg], ClientState) {- ^ response , updated state -}
-applyMessageToClientState time irc networkId cs st =
-  cs' `seq` (reply, st')
+  ([RawIrcMsg], Maybe DCCUpdate, ClientState) {- ^ response , DCC updates, updated state -}
+applyMessageToClientState time irc network cs st =
+  cs' `seq` (reply, dccUp, st')
   where
     (reply, cs') = applyMessage time irc cs
-    network      = view csNetwork cs
-    st'          = applyWindowRenames network irc
-                 $ set (clientConnections . ix networkId) cs' st
+    (st', dccUp) = queueDCCTransfer network irc
+                 $ applyWindowRenames network irc
+                 $ set (clientConnections . ix network) cs' st
+
+-- | Queue a DCC transfer when the message is correct. Await for user
+--   confirmation to start the download.
+queueDCCTransfer :: Text -> IrcMsg -> ClientState
+                 -> (ClientState, Maybe DCCUpdate)
+queueDCCTransfer network ctcpMsg st
+  | Just (fromU, _target, command, txt) <- ctcpToTuple ctcpMsg
+  , command == "DCC", dccState <- view clientDCC st
+  = case (parseSEND network fromU txt, parseACCEPT dccState fromU txt) of
+      (Right offer, _) -> (set clientDCC (insertAsNewMax offer dccState) st, Nothing)
+      (_, Just upd) -> (st, Just upd)
+      (_, _) -> (st, Nothing)
+
+  | otherwise = (st, Nothing)
+
 
 -- | When a nick change happens and there is an open query window for that nick
 -- and there isn't an open query window for the new nick, rename the window.
@@ -784,45 +815,6 @@ applyWindowRenames network (Nick old new) st
 applyWindowRenames _ _ st = st
 
 
--- | Actions to be run when exiting the client.
-clientShutdown :: ClientState -> IO ()
-clientShutdown st = () <$ clientStopExtensions st
- -- other shutdown stuff might be added here later
-
-
--- | Unload all active extensions.
-clientStopExtensions :: ClientState -> IO ClientState
-clientStopExtensions st =
-  do let (aes,st1) = (clientExtensions . esActive <<.~ []) st
-     (st2,_) <- clientPark st1 $ \ptr ->
-                  traverse_ (deactivateExtension ptr) aes
-     return st2
-
--- | Start extensions after ensuring existing ones are stopped
-clientStartExtensions :: ClientState -> IO ClientState
-clientStartExtensions st =
-  do let cfg = view clientConfig st
-     st1        <- clientStopExtensions st
-     (st2, res) <- clientPark st1 $ \ptr ->
-            traverse (try . activateExtension ptr)
-                     (view configExtensions cfg)
-
-     let (errors, exts) = partitionEithers res
-     st3 <- recordErrors errors st2
-     return $! set (clientExtensions . esActive) exts st3
-  where
-    recordErrors [] ste = return ste
-    recordErrors es ste =
-      do now <- getZonedTime
-         return $! foldl' (recordError now) ste es
-
-    recordError now ste e =
-      recordNetworkMessage ClientMessage
-        { _msgTime    = now
-        , _msgBody    = ErrorBody (Text.pack (show (e :: IOError)))
-        , _msgNetwork = ""
-        } ste
-
 ------------------------------------------------------------------------
 -- Scrolling
 ------------------------------------------------------------------------
@@ -848,15 +840,17 @@ clientExtraFocuses st =
 -- Some events like errors or chat messages mentioning keywords are
 -- considered important and will be jumped to first.
 jumpToActivity :: ClientState -> ClientState
-jumpToActivity st = changeFocus newFocus st
+jumpToActivity st =
+  case mplus highPriority lowPriority of
+    Just (focus,_) -> changeFocus focus st
+    Nothing ->
+      case view clientActivityReturn st of
+        Just focus -> changeFocus focus st
+        Nothing    -> st
   where
     windowList   = views clientWindows Map.toAscList st
     highPriority = find (\x -> WLImportant == view winMention (snd x)) windowList
     lowPriority  = find (\x -> view winUnread (snd x) > 0) windowList
-    newFocus =
-      case mplus highPriority lowPriority of
-        Just (focus,_) -> focus
-        Nothing        -> view clientActivityReturn st
 
 -- | Jump the focus directly to a window based on its zero-based index.
 jumpFocus ::
