@@ -31,22 +31,27 @@ module Client.State.Network
   , csChannelTypes
   , csTransaction
   , csModes
+  , csSnomask
   , csStatusMsg
   , csSettings
   , csUserInfo
   , csUsers
   , csUser
   , csModeCount
-  , csNetworkId
   , csNetwork
   , csNextPingTime
   , csPingStatus
+  , csLatency
   , csLastReceived
+  , csCertificate
   , csMessageHooks
   , csAuthenticationState
 
   -- * User information
   , UserAndHost(..)
+  , uhUser
+  , uhHost
+  , uhAccount
 
   -- * Cross-message state
   , Transaction(..)
@@ -106,24 +111,29 @@ import           LensUtils
 
 -- | State tracked for each IRC connection
 data NetworkState = NetworkState
-  { _csNetworkId    :: !NetworkId -- ^ network connection identifier
-  , _csChannels     :: !(HashMap Identifier ChannelState) -- ^ joined channels
+  { _csChannels     :: !(HashMap Identifier ChannelState) -- ^ joined channels
   , _csSocket       :: !NetworkConnection -- ^ network socket
   , _csModeTypes    :: !ModeTypes -- ^ channel mode meanings
+  , _csUmodeTypes   :: !ModeTypes -- ^ user mode meanings
   , _csChannelTypes :: ![Char] -- ^ channel identifier prefixes
   , _csTransaction  :: !Transaction -- ^ state for multi-message sequences
   , _csModes        :: ![Char] -- ^ modes for the connected user
+  , _csSnomask      :: ![Char] -- ^ server notice modes for the connected user
   , _csStatusMsg    :: ![Char] -- ^ modes that prefix statusmsg channel names
   , _csSettings     :: !ServerSettings -- ^ settings used for this connection
   , _csUserInfo     :: !UserInfo -- ^ usermask used by the server for this connection
   , _csUsers        :: !(HashMap Identifier UserAndHost) -- ^ user and hostname for other nicks
   , _csModeCount    :: !Int -- ^ maximum mode changes per MODE command
   , _csNetwork      :: !Text -- ^ name of network connection
-  , _csNextPingTime :: !(Maybe UTCTime) -- ^ time for next ping event
-  , _csPingStatus   :: !PingStatus -- ^ state of ping timer
-  , _csLastReceived :: !(Maybe UTCTime) -- ^ time of last message received
   , _csMessageHooks :: ![Text] -- ^ names of message hooks to apply to this connection
   , _csAuthenticationState :: !AuthenticateState
+
+  -- Timing information
+  , _csNextPingTime :: !(Maybe UTCTime) -- ^ time for next ping event
+  , _csLatency      :: !(Maybe NominalDiffTime) -- ^ latency calculated from previous pong
+  , _csPingStatus   :: !PingStatus      -- ^ state of ping timer
+  , _csLastReceived :: !(Maybe UTCTime) -- ^ time of last message received
+  , _csCertificate  :: ![Text]
   }
   deriving Show
 
@@ -133,33 +143,45 @@ data AuthenticateState
   | AS_PlainStarted       -- ^ PLAIN mode initiated
   | AS_EcdsaStarted       -- ^ ECDSA-NIST mode initiated
   | AS_EcdsaWaitChallenge -- ^ ECDSA-NIST user sent waiting for challenge
+  | AS_ExternalStarted    -- ^ EXTERNAL mode initiated
   deriving Show
 
 -- | Pair of username and hostname. Empty strings represent missing information.
-data UserAndHost =
-  UserAndHost {-# UNPACK #-} !Text {-# UNPACK #-} !Text
-  -- ^ username hostname
+data UserAndHost = UserAndHost
+  { _uhUser    :: {-# UNPACK #-}!Text -- ^ username
+  , _uhHost    :: {-# UNPACK #-}!Text -- ^ hostname
+  , _uhAccount :: {-# UNPACK #-}!Text -- ^ services account
+  }
   deriving Show
 
 -- | Status of the ping timer
 data PingStatus
-  = PingSent    !UTCTime -- ^ ping sent waiting for pong
-  | PingLatency !Double -- ^ latency in seconds for last ping
-  | PingNever -- ^ no ping sent
+  = PingSent !UTCTime -- ^ ping sent at given time, waiting for pong
+  | PingNone          -- ^ not waiting for a pong
   | PingConnecting !Int !(Maybe UTCTime) -- ^ number of attempts, last known connection time
   deriving Show
 
+-- | Timer-based events
+data TimedAction
+  = TimedDisconnect    -- ^ terminate the connection due to timeout
+  | TimedSendPing      -- ^ transmit a ping to the server
+  | TimedForgetLatency -- ^ erase latency (when it is outdated)
+  deriving (Eq, Ord, Show)
 
 data Transaction
   = NoTransaction
   | NamesTransaction [Text]
   | BanTransaction [(Text,MaskListEntry)]
   | WhoTransaction [UserInfo]
+  | CapTransaction
+  | CapLsTransaction [(Text, Maybe Text)]
   deriving Show
 
 makeLenses ''NetworkState
+makeLenses ''UserAndHost
 makePrisms ''Transaction
 makePrisms ''PingStatus
+makePrisms ''TimedAction
 
 defaultChannelTypes :: String
 defaultChannelTypes = "#&"
@@ -179,12 +201,26 @@ sendMsg cs msg =
   where
     transmit = send (view csSocket cs) . renderRawIrcMsg
 
+    actionPrefix = "\^AACTION "
+    actionSuffix = "\^A"
+
+    -- Special case for splitting a single CTCP ACTION into
+    -- multiple actions
+    multiline cmd tgt txt
+      | Just txt1 <- Text.stripPrefix actionPrefix txt
+      , Just txt2 <- Text.stripSuffix actionSuffix txt1 =
+      let txtChunks     = utf8ChunksOf maxContentLen txt2
+          maxContentLen = computeMaxMessageLength (view csUserInfo cs) tgt
+                        - Text.length actionPrefix - Text.length actionSuffix
+      in for_ txtChunks $ \txtChunk ->
+           transmit $ rawIrcMsg cmd [tgt, actionPrefix <> txtChunk <> actionSuffix]
+
+    -- Normal case
     multiline cmd tgt txt =
-      for_ txtChunks $ \txtChunk ->
-        transmit $ rawIrcMsg cmd [tgt, txtChunk]
-      where
-        txtChunks = utf8ChunksOf maxContentLen txt
-        maxContentLen = computeMaxMessageLength (view csUserInfo cs) tgt
+      let txtChunks     = utf8ChunksOf maxContentLen txt
+          maxContentLen = computeMaxMessageLength (view csUserInfo cs) tgt
+      in for_ txtChunks $ \txtChunk ->
+           transmit $ rawIrcMsg cmd [tgt, txtChunk]
 
 -- This is an approximation for splitting the text. It doesn't
 -- understand combining characters. A correct implementation
@@ -215,31 +251,33 @@ utf8ChunksOf n txt
 -- | Construct a new network state using the given settings and
 -- default values as specified by the IRC specification.
 newNetworkState ::
-  NetworkId         {- ^ unique network ID         -} ->
   Text              {- ^ network name              -} ->
   ServerSettings    {- ^ server settings           -} ->
   NetworkConnection {- ^ active network connection -} ->
   PingStatus        {- ^ initial ping status       -} ->
   NetworkState      {- ^ new network state         -}
-newNetworkState networkId network settings sock ping = NetworkState
-  { _csNetworkId    = networkId
-  , _csUserInfo     = UserInfo "*" "" ""
+newNetworkState network settings sock ping = NetworkState
+  { _csUserInfo     = UserInfo "*" "" ""
   , _csChannels     = HashMap.empty
   , _csSocket       = sock
   , _csChannelTypes = defaultChannelTypes
   , _csModeTypes    = defaultModeTypes
-  , _csTransaction  = NoTransaction
+  , _csUmodeTypes   = defaultUmodeTypes
+  , _csTransaction  = CapTransaction
   , _csModes        = ""
+  , _csSnomask      = ""
   , _csStatusMsg    = ""
   , _csSettings     = settings
   , _csModeCount    = 3
   , _csUsers        = HashMap.empty
   , _csNetwork      = network
-  , _csPingStatus   = ping
-  , _csNextPingTime = Nothing
-  , _csLastReceived = Nothing
   , _csMessageHooks = view ssMessageHooks settings
   , _csAuthenticationState = AS_None
+  , _csPingStatus   = ping
+  , _csLatency      = Nothing
+  , _csNextPingTime = Nothing
+  , _csLastReceived = Nothing
+  , _csCertificate  = []
   }
 
 
@@ -265,12 +303,20 @@ applyMessage' msgWhen msg cs =
   case msg of
     Ping args -> ([ircPong args], cs)
     Pong _    -> noReply $ doPong msgWhen cs
-    Join user chan ->
+    Join user chan acct ->
            noReply
-         $ recordUser user
+         $ recordUser user acct
          $ overChannel chan (joinChannel (userNick user))
          $ createOnJoin user chan cs
 
+    Account user acct ->
+           noReply
+         $ recordUser user acct cs
+
+    Chghost user newUser newHost ->
+           noReply
+         $ updateUserInfo (userNick user) newUser newHost cs
+
     Quit user _reason ->
            noReply
          $ forgetUser (userNick user)
@@ -287,12 +333,20 @@ applyMessage' msgWhen msg cs =
          $ overChannels (nickChange (userNick oldNick) newNick) cs
 
     Reply RPL_WELCOME (me:_) -> doWelcome msgWhen (mkId me) cs
-    Reply RPL_SASLSUCCESS _ -> ([ircCapEnd], cs)
-    Reply RPL_SASLFAIL _ -> ([ircCapEnd], cs)
+    Reply RPL_SASLSUCCESS _ -> endCapTransaction cs
+    Reply RPL_SASLFAIL _ -> endCapTransaction cs
     Reply ERR_NICKNAMEINUSE (_:badnick:_)
       | PingConnecting{} <- view csPingStatus cs -> doBadNick badnick cs
+    Reply RPL_HOSTHIDDEN (_:host:_) ->
+        noReply (set (csUserInfo . uiHost) host cs)
+
+    -- /who <#channel> %tuhna,616
+    Reply RPL_WHOSPCRPL [_me,"616",user,host,nick,acct] ->
+       let acct' = if acct == "0" then "" else acct
+       in noReply (recordUser (UserInfo (mkId nick) user host) acct' cs)
+
     Reply code args        -> noReply (doRpl code msgWhen args cs)
-    Cap cmd params         -> doCap cmd params cs
+    Cap cmd                -> doCap cmd cs
     Authenticate param     -> doAuthenticate param cs
     Mode who target (modes:params)  -> doMode msgWhen who target modes params cs
     Topic user chan topic  -> noReply (doTopic msgWhen user chan topic cs)
@@ -324,7 +378,8 @@ doWelcome msgWhen me
   = noReply
   . set csNick me
   . set csNextPingTime (Just $! addUTCTime 30 (zonedTimeToUTC msgWhen))
-  . set csPingStatus PingNever
+  . set csPingStatus PingNone
+  . set csTransaction NoTransaction -- wipe out CapTransaction if it was active
 
 -- | Handle 'ERR_NICKNAMEINUSE' errors when connecting.
 doBadNick ::
@@ -356,12 +411,20 @@ doRpl :: ReplyCode -> ZonedTime -> [Text] -> NetworkState -> NetworkState
 doRpl cmd msgWhen args =
   case cmd of
     RPL_UMODEIS ->
+      \ns ->
       case args of
         _me:modes:params
-          | Just xs <- splitModes defaultUmodeTypes modes params ->
+          | Just xs <- splitModes (view csUmodeTypes ns) modes params ->
                  doMyModes xs
-               . set csModes "" -- reset modes
-        _ -> id
+               $ set csModes "" ns -- reset modes
+        _ -> ns
+
+    RPL_SNOMASK ->
+      case args of
+        _me:snomask0:_
+          | Just snomask <- Text.stripPrefix "+" snomask0 ->
+           set csSnomask (Text.unpack snomask)
+        _             -> id
 
     RPL_NOTOPIC ->
       case args of
@@ -397,6 +460,8 @@ doRpl cmd msgWhen args =
           overChannel (mkId chan) (set chanUrl (Just urlTxt))
         _ -> id
 
+    RPL_MYINFO -> myinfo args
+
     RPL_ISUPPORT -> isupport args
 
     RPL_NAMREPLY ->
@@ -526,14 +591,15 @@ squelchReply rpl =
     RPL_QUIETLIST       -> True
     RPL_ENDOFQUIETLIST  -> True
     RPL_CHANNELMODEIS   -> True
-    RPL_TOPIC           -> True
+    RPL_UMODEIS         -> True
+    RPL_SNOMASK         -> True
+    RPL_WHOREPLY        -> True
+    RPL_ENDOFWHO        -> True
+    RPL_WHOSPCRPL       -> True
     RPL_TOPICWHOTIME    -> True
     RPL_CREATIONTIME    -> True
     RPL_CHANNEL_URL     -> True
     RPL_NOTOPIC         -> True
-    RPL_UMODEIS         -> True
-    RPL_WHOREPLY        -> True
-    RPL_ENDOFWHO        -> True
     _                   -> False
 
 -- | Return 'True' for messages that should be hidden outside of
@@ -553,7 +619,7 @@ doMode ::
   NetworkState -> ([RawIrcMsg], NetworkState)
 doMode when who target modes args cs
   | view csNick cs == target
-  , Just xs <- splitModes defaultUmodeTypes modes args =
+  , Just xs <- splitModes (view csUmodeTypes cs) modes args =
         noReply (doMyModes xs cs)
 
   | isChannelIdentifier cs target
@@ -621,30 +687,53 @@ doMyModes changes = over csModes $ \modes -> sort (foldl' applyOne modes changes
       | otherwise         = mode:modes
     applyOne modes (False, mode, _) = delete mode modes
 
-supportedCaps :: NetworkState -> [Text]
-supportedCaps cs =
-  sasl ++ ["multi-prefix", "znc.in/batch", "znc.in/playback",
-           "znc.in/server-time-iso", "znc.in/self-message"]
+selectCaps ::
+  NetworkState         {- ^ network state  -} ->
+  [(Text, Maybe Text)] {- ^ server caps    -} ->
+  [Text]               {- ^ caps to enable -}
+selectCaps cs offered = supported `intersect` Map.keys capMap
   where
+    capMap = Map.fromList offered
+
+    supported =
+      sasl ++ serverTime ++
+      ["multi-prefix", "batch", "znc.in/playback", "znc.in/self-message"
+      , "cap-notify", "extended-join", "account-notify", "chghost"
+      , "userhost-in-names", "account-tag" ]
+
+    -- logic for using IRCv3.2 server-time if available and falling back
+    -- to ZNC's specific extension otherwise.
+    serverTime
+      | "server-time"            `Map.member` capMap = ["server-time"]
+      | "znc.in/server-time-iso" `Map.member` capMap = ["znc.in/server-time-iso"]
+      | otherwise                                    = []
+
     ss = view csSettings cs
     sasl = ["sasl" | isJust (view ssSaslUsername ss)
                    , isJust (view ssSaslPassword ss) ||
-                     isJust (view ssSaslEcdsaFile ss) ]
+                     isJust (view ssSaslEcdsaFile ss) ||
+                     isJust (view ssTlsClientCert ss) ]
 
 doAuthenticate :: Text -> NetworkState -> ([RawIrcMsg], NetworkState)
 doAuthenticate param cs =
   case view csAuthenticationState cs of
     AS_PlainStarted
-      | "+" <- param
+      | "+"       <- param
       , Just user <- view ssSaslUsername ss
       , Just pass <- view ssSaslPassword ss
-      -> ([ircAuthenticate (encodePlainAuthentication user pass)],
+      -> (ircAuthenticates (encodePlainAuthentication user pass),
+          set csAuthenticationState AS_None cs)
+
+    AS_ExternalStarted
+      | "+"       <- param
+      , Just user <- view ssSaslUsername ss
+      -> (ircAuthenticates (encodeExternalAuthentication user),
           set csAuthenticationState AS_None cs)
 
     AS_EcdsaStarted
-      | "+" <- param
+      | "+"       <- param
       , Just user <- view ssSaslUsername ss
-      -> ([ircAuthenticate (Ecdsa.encodeUsername user)],
+      -> (ircAuthenticates (Ecdsa.encodeUsername user),
           set csAuthenticationState AS_EcdsaWaitChallenge cs)
 
     AS_EcdsaWaitChallenge -> ([], cs) -- handled in Client.EventLoop!
@@ -655,36 +744,57 @@ doAuthenticate param cs =
     ss = view csSettings cs
 
 
-doCap :: CapCmd -> [Text] -> NetworkState -> ([RawIrcMsg], NetworkState)
-doCap cmd args cs =
-  case (cmd,args) of
-    (CapLs,[capsTxt])
-      | null reqCaps -> ([ircCapEnd], cs)
-      | otherwise -> ([ircCapReq reqCaps], cs)
+doCap :: CapCmd -> NetworkState -> ([RawIrcMsg], NetworkState)
+doCap cmd cs =
+  case cmd of
+    (CapLs CapMore caps) ->
+      noReply (set csTransaction (CapLsTransaction (caps ++ prevCaps)) cs)
       where
-        caps = Text.words capsTxt
-        reqCaps = intersect (supportedCaps cs) caps
+        prevCaps = view (csTransaction . _CapLsTransaction) cs
 
-    (CapAck,[capsTxt])
-      | "sasl" `elem` caps ->
+    CapLs CapDone caps
+      | null reqCaps -> endCapTransaction cs
+      | otherwise    -> ([ircCapReq reqCaps], cs)
+      where
+        reqCaps = selectCaps cs (caps ++ view (csTransaction . _CapLsTransaction) cs)
+
+    CapNew caps
+      | null reqCaps -> ([], cs)
+      | otherwise    -> ([ircCapReq reqCaps], cs)
+      where
+        reqCaps = selectCaps cs caps
+
+    CapDel _ -> ([],cs)
+
+    CapAck caps
+      | "sasl" `elem` caps && isJust (view ssSaslUsername ss) ->
           if isJust (view ssSaslEcdsaFile ss)
-            then ([ircAuthenticate Ecdsa.authenticationMode], set csAuthenticationState AS_EcdsaStarted cs)
-          else if isJust (view ssSaslUsername ss)
-            then ([ircAuthenticate plainAuthenticationMode], set csAuthenticationState AS_PlainStarted cs)
-          else ([ircCapEnd], cs)
+            then ( [ircAuthenticate Ecdsa.authenticationMode]
+                 , set csAuthenticationState AS_EcdsaStarted cs)
+          else if isJust (view ssSaslPassword ss)
+            then ( [ircAuthenticate "PLAIN"]
+                 , set csAuthenticationState AS_PlainStarted cs)
+          else if isJust (view ssTlsClientCert ss)
+            then ( [ircAuthenticate "EXTERNAL"]
+                 , set csAuthenticationState AS_ExternalStarted cs)
+          else endCapTransaction cs
       where
-        ss   = view csSettings cs
-        caps = Text.words capsTxt
+        ss = view csSettings cs
 
-    _ -> ([ircCapEnd], cs)
+    _ -> endCapTransaction cs
 
+endCapTransaction :: NetworkState -> ([RawIrcMsg], NetworkState)
+endCapTransaction cs =
+  case view csTransaction cs of
+    CapTransaction -> ([ircCapEnd], set csTransaction NoTransaction cs)
+    _              -> ([], cs)
 
 initialMessages :: NetworkState -> [RawIrcMsg]
 initialMessages cs
    = [ ircCapLs ]
   ++ [ ircPass pass | Just pass <- [view ssPassword ss]]
   ++ [ ircNick (views ssNicks NonEmpty.head ss)
-     , ircUser (view ssUser ss) False True (view ssReal ss)
+     , ircUser (view ssUser ss) (view ssReal ss)
      ]
   where
     ss = view csSettings cs
@@ -692,20 +802,29 @@ initialMessages cs
 loadNamesList :: Identifier -> NetworkState -> NetworkState
 loadNamesList chan cs
   = set csTransaction NoTransaction
+  $ flip (foldl' (flip learnUserInfo)) (fst <$> entries)
   $ setStrict (csChannels . ix chan . chanUsers) newChanUsers
   $ cs
   where
-    newChanUsers = HashMap.fromList (splitEntry "" <$> entries)
+    newChanUsers = HashMap.fromList [ (view uiNick ui, modes) | (ui, modes) <- entries ]
+
+    -- userhost-in-names might or might not include the user and host
+    -- if we find it we update the user information.
+    learnUserInfo (UserInfo n u h)
+      | Text.null u || Text.null h = id
+      | otherwise = updateUserInfo n u h
 
     sigils = toListOf (csModeTypes . modesPrefixModes . folded . _2) cs
 
     splitEntry modes str
       | Text.head str `elem` sigils = splitEntry (Text.head str : modes)
                                                  (Text.tail str)
-      | otherwise = (mkId str, reverse modes)
+      | otherwise = (parseUserInfo str, reverse modes)
 
-    entries =
-      concatMap Text.words (view (csTransaction . _NamesTransaction) cs)
+    entries :: [(UserInfo, [Char])]
+    entries = fmap (splitEntry "")
+            $ concatMap Text.words
+            $ view (csTransaction . _NamesTransaction) cs
 
 
 createOnJoin :: UserInfo -> Identifier -> NetworkState -> NetworkState
@@ -720,6 +839,15 @@ updateMyNick oldNick newNick cs
   | oldNick == view csNick cs = set csNick newNick cs
   | otherwise = cs
 
+myinfo ::
+  [Text] ->
+  NetworkState ->
+  NetworkState
+myinfo (_me : _host : _version : umodes : _) =
+  -- special logic for s because I know it has arguments
+  set (csUmodeTypes . modesNeverArg) (delete 's' (Text.unpack umodes))
+myinfo _ = id
+
 -- ISUPPORT is defined by
 -- https://tools.ietf.org/html/draft-brocklesby-irc-isupport-03#section-3.14
 isupport ::
@@ -790,11 +918,21 @@ isChannelIdentifier cs ident =
 csUser :: Functor f => Identifier -> LensLike' f NetworkState (Maybe UserAndHost)
 csUser i = csUsers . at i
 
-recordUser :: UserInfo -> NetworkState -> NetworkState
-recordUser (UserInfo nick user host)
+recordUser :: UserInfo -> Text -> NetworkState -> NetworkState
+recordUser (UserInfo nick user host) acct
   | Text.null user || Text.null host = id
   | otherwise = set (csUsers . at nick)
-                    (Just (UserAndHost user host))
+                    (Just $! UserAndHost user host acct)
+
+-- | Process a CHGHOST command, updating a users information
+updateUserInfo ::
+  Identifier {- ^ nickname     -} ->
+  Text       {- ^ new username -} ->
+  Text       {- ^ new hostname -} ->
+  NetworkState -> NetworkState
+updateUserInfo nick user host =
+  over (csUsers . at nick) $ \old ->
+    Just $! UserAndHost user host (maybe "" _uhAccount old)
 
 forgetUser :: Identifier -> NetworkState -> NetworkState
 forgetUser = over csUsers . sans
@@ -828,40 +966,60 @@ massRegistration cs
       | not (Text.null user)
       , not (Text.null host)
       , HashSet.member nick channelUsers =
-              HashMap.insert nick (UserAndHost user host) users
+            HashMap.alter
+                (\mb -> case mb of
+                          Nothing                     -> Just $! UserAndHost user host ""
+                          Just (UserAndHost _ _ acct) -> Just $! UserAndHost user host acct
+                ) nick users
       | otherwise = users
 
--- | Timer-based events
-data TimedAction
-  = TimedDisconnect -- ^ terminate the connection due to timeout
-  | TimedSendPing -- ^ transmit a ping to the server
-  deriving (Eq, Ord, Show)
-
 -- | Compute the earliest timed action for a connection, if any
 nextTimedAction :: NetworkState -> Maybe (UTCTime, TimedAction)
-nextTimedAction cs =
+nextTimedAction ns = minimumOf (folded.folded) actions
+  where
+    actions = [nextPingAction ns, nextForgetAction ns]
+
+-- | Compute the timed action for forgetting the ping latency.
+-- The client will wait for a multiple of the current latency
+-- for the next pong response in order to reduce jitter in
+-- the rendered latency when everything is fine.
+nextForgetAction :: NetworkState -> Maybe (UTCTime, TimedAction)
+nextForgetAction ns =
+  do sentAt  <- preview (csPingStatus . _PingSent) ns
+     latency <- view csLatency ns
+     let delay = max 0.1 (3 * latency) -- wait at least 0.1s (ensure positive waits)
+         eventAt = addUTCTime delay sentAt
+     return (eventAt, TimedForgetLatency)
+
+-- | Compute the next action needed for the client ping logic.
+nextPingAction :: NetworkState -> Maybe (UTCTime, TimedAction)
+nextPingAction cs =
   do runAt <- view csNextPingTime cs
      return (runAt, action)
   where
     action =
       case view csPingStatus cs of
-        PingSent{}    -> TimedDisconnect
-        PingLatency{} -> TimedSendPing
-        PingNever     -> TimedSendPing
+        PingSent{}       -> TimedDisconnect
+        PingNone         -> TimedSendPing
         PingConnecting{} -> TimedSendPing
 
 doPong :: ZonedTime -> NetworkState -> NetworkState
-doPong when cs = set csPingStatus (PingLatency delta) cs
+doPong when cs = set csPingStatus PingNone
+               $ set csLatency (Just delta) cs
   where
     delta =
       case view csPingStatus cs of
-        PingSent sent -> realToFrac (diffUTCTime (zonedTimeToUTC when) sent)
+        PingSent sent -> diffUTCTime (zonedTimeToUTC when) sent
         _             -> 0
 
 -- | Apply the given 'TimedAction' to a connection state.
 applyTimedAction :: TimedAction -> NetworkState -> IO NetworkState
 applyTimedAction action cs =
+
   case action of
+    TimedForgetLatency ->
+      do return $! set csLatency Nothing cs
+
     TimedDisconnect ->
       do abortConnection PingTimeout (view csSocket cs)
          return $! set csNextPingTime Nothing cs
