@@ -28,10 +28,10 @@ import           Client.State.Window
 import           Control.Lens
 import           Control.Monad
 import           Data.List
-import           Data.Semigroup
 import           Irc.Identifier
 import           Irc.Message
 import           Irc.UserInfo
+import           ContextFilter (filterContext)
 
 
 chatMessageImages :: Focus -> Int -> ClientState -> [Image']
@@ -42,7 +42,7 @@ chatMessageImages focus w st =
       let msgs     = toListOf each (view winMessages win)
           hideMeta = view winHideMeta win in
       case clientMatcher st of
-        Just matcher -> windowLineProcessor hideMeta (filter (views wlText matcher) msgs)
+        Just (Matcher a b p) -> windowLineProcessor hideMeta (filterContext b a (views wlText p) msgs)
         Nothing ->
           case view winMarker win of
             Nothing -> windowLineProcessor hideMeta msgs
@@ -171,4 +171,5 @@ metadataWindowLine ::
 metadataWindowLine st wl =
   case view wlSummary wl of
     ChatSummary who -> (ignoreImage, userNick who, Nothing) <$ guard (identIgnored who st)
+    DccSendSummary _ -> Nothing -- Show a custom WindowLine latter
     summary         -> metadataImg summary
