@@ -23,9 +23,7 @@ import           Control.Lens
 import qualified Data.HashMap.Strict as HashMap
 import qualified Data.Map.Strict as Map
 import           Data.List
-import           Data.Maybe
 import           Data.Ord
-import           Data.Semigroup
 import           Data.Text (Text)
 import qualified Data.Text as Text
 import qualified Data.Text.Lazy as LText
@@ -54,7 +52,7 @@ userListImages' cs channel st =
   where
     countImage = drawSigilCount pal (map snd usersList)
 
-    matcher = fromMaybe (const True) (clientMatcher st)
+    matcher = maybe (const True) matcherPred (clientMatcher st)
 
     myNicks = clientHighlights cs st
 
@@ -107,7 +105,7 @@ userInfoImages network channel st =
 userInfoImages' :: NetworkState -> Identifier -> ClientState -> [Image']
 userInfoImages' cs channel st = countImage : map renderEntry usersList
   where
-    matcher = fromMaybe (const True) (clientMatcher st)
+    matcher = maybe (const True) matcherPred (clientMatcher st)
 
     countImage = drawSigilCount pal (map snd usersList)
 
@@ -115,21 +113,22 @@ userInfoImages' cs channel st = countImage : map renderEntry usersList
 
     pal = clientPalette st
 
-    renderEntry (info, sigils) =
+    renderEntry ((info, acct), sigils) =
       string (view palSigil pal) sigils <>
-      coloredUserInfo pal DetailedRender myNicks info
+      coloredUserInfo pal DetailedRender myNicks info <>
+      " " <> text' (view palMeta pal) (cleanText acct)
 
-    matcher' (info,sigils) =
-      matcher (LText.fromChunks [Text.pack sigils, renderUserInfo info])
+    matcher' ((info, acct),sigils) =
+      matcher (LText.fromChunks [Text.pack sigils, renderUserInfo info, " ", acct])
 
     userInfos = view csUsers cs
 
     toInfo nick =
       case view (at nick) userInfos of
-        Just (UserAndHost n h) -> UserInfo nick n h
-        Nothing                -> UserInfo nick "" ""
+        Just (UserAndHost n h a) -> (UserInfo nick n h, a)
+        Nothing                  -> (UserInfo nick "" "", "")
 
-    usersList = sortBy (flip (comparing (userNick . fst)))
+    usersList = sortBy (flip (comparing (userNick . fst . fst)))
               $ filter matcher'
               $ map (over _1 toInfo)
               $ HashMap.toList usersHashMap
